unit FABTIImageReader;
// ***********************************************************************
// *                                                                     *
// *   TI Diskette Image/File Reader - Reformatter                       *
// *   RGM, The summer of Nostalgia, 2003                                *
// *                                                                     *
// *   This program will handle files that have been transferred to the  *                                                     *
// *   PC using XMODEM or any other protocol which provides a TIFILES    *
// *   header. It will also read a V9T9 diskette image and display       *
// *   individual files, and extract them constructing an appropriate    *
// *   TIHEADERs.                                                        *
// *                                                                     *
// *   Formatting of TI-Writer and 'companion' files is rudimentary, but *
// *   does clean ip formatting codes and leave a clean block of text,   *
// *   line-oriented in the case of TI Writer and paragraph-oriented     *
// *   (a-la unwrapped notepad texts) in the case of 'companion.'        *
// *   Formatted output can be saved as plain text or as an RFT.         *
// *                                                                     *
// ***********************************************************************

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls,
  //
  lmdstdcS, lmdctrl, lmdextcS, lmdcctrl, Menus, ActnList, ImgList, lmdcompo,
  lmdclass, lmddlgS, VersInfo, ComCtrls, ToolWin, LMDCustomComponent,
  LMDContainerComponent, LMDBaseDialog, LMDAboutDlg, LMDCustomSimpleLabel,
  LMDSimpleLabel, LMDControl, LMDBaseControl, LMDBaseGraphicControl,
  LMDBaseLabel, LMDInformationLabel, LMDCustomControl, LMDCustomPanel,
  LMDCustomBevelPanel, LMDCustomStatusBar, LMDStatusBar, StdActns,
  ExtActns, ActnMan, Grids, DBGrids, RXDBCtrl, CheckRxDBGrid, DB, DBTables,
  Spin, Numedit, AdvGrid, AdvCGrid, ActnCtrls, RxCombos, RXSpin, ProDlgs,
  Printers, LMDCustomScrollBox, LMDListBox, ActnMenus, RXGrids,
  RXCtrls, DBXpress, Provider, SqlExpr, DBClient, DBLocal, DBLocalS,
  LMDBaseEdit, LMDCustomEdit, LMDCustomBrowseEdit, LMDBrowseEdit,
  LMDCustomComboBox, LMDDriveComboBox, ShellCtrls, FileCtrl, RxRichEd;

type
  TTIFilesHeader = Record
    HeaderString: string[7]; // bytes 1 - 7
    Sectors: integer;        // bytes 8 - 9
    StatusFlag: integer;     // byte 11
    //
    FileType: string;        // interpreted from Byte 10 and RecSize (Byte 13)
    //
    //  a more detailed interpretation of the StatusFlag
    //
    ProgType: boolean;    // IMAGE (Program) if TRUE
    DataSize: string[3];  // VAR or FIX
    DataType: string[3];  // DISplay or INTernal (7 or 8 bit)
    ProtectBit : boolean; // Protected if TRUE
    //
    RecordsPerSec: integer;  // byte 11
    EOFOffset: integer;      // byte 12
    RecSize: integer;        // byte 13
    RecCount: integer;       // bytes 14-15 in theory but really only byte 14?
                             //    FIX files repeat SECTOR count
                             //    VAR Files it is probably the true record count
                             //      possibly with bytes 15 and 14 reversed for
                             //      counts > 255
    end;

  TTISector = Record
    EndOffset: integer;                   //recorded but never used?
    Contents: array[0..255] of char;
    end;
                                   
  TTISegmentGroup = Record         // This is a single entry in the segmet list
    Cluster: array[0..2] of char;  //  identifying the starting segment and the

    StartSeg: integer;             //  segment counter in nybbles >AB >CD >EF

    EndOffset: integer;            //  where >DAB is the sector number and

    end;                           //        >EFC is the "number of sectors -1

                                   //  in this segment added to the number of

                                   //  sectors in all previous segments" per

                                   //  DiskFixer. So, if this number is 20 and

                                   //  the number in the previous entry was 18,

                                   //  there are 2 sectors in this segment.


  TTISegmentList = array[0..74] of TTISegmentGroup;   //just set at max count

  TTIFileDescriptorRecord = Record
    Address: integer;               // sector number
    Filename: string[10];           // bytes 0 - 9
    StatusFlag: integer;              // byte 12
    RecordsPerSector: integer;      // byte 13
    SectorCount: integer;           // bytes 14-15
    LastByteInLastSector: integer;  // byte 16
    Recordlength: integer;          // byte 17
    NumberRecordsSectors: integer;  // bytes 18 and 19 but swapped!
    SegmentList: TTISegmentList;
    end;

  TTIDiskInformation =  Record
    Diskname: string[10];       // bytes 0 - 9
    SectorCount: integer;       // bytes 10 and 11
    SectorsPerTrack: integer;   // byte 12
    DSRMark: string[3];         // bytes 13 - 15
    DiskProtected: boolean;     // byte 16
    TracksPerSide: integer;     // byte 17
    Sides: integer;             // byte 18
    Density: integer;           // byte 19
    Bitmap: array of char;      // starts at byte $38
    FileCount: integer;         // # of FDR pointers in sec #1
    end;

  TTIDisk = Record
    PCName: TFilename;
    Info: TTIDiskInformation;
    SectorList: array of TTISector;
    SectorsRead: boolean;
    Directory: array of TTIFileDescriptorRecord;
    end;

  TTIFile = Record              // same format for extract or stand-alone
    Name: string;
    PCName: TFilename;
    Header: TTIFilesHeader;
    SectorList: array of TTISector;
    StandAlone: boolean;
    SectorsRead: boolean;
    BasicFlag: boolean;
    CompanionFlag: boolean;
    TIWriterFlag: boolean;
    FormattedFlag: boolean;
    end;

  TfrmMain = class(TForm)
    statMain: TLMDStatusBar;
    statMainDateTime: TLMDInformationLabel;
    statMainVariable: TLMDInformationLabel;
    statMainHint: TLMDSimpleLabel;
    pnlTopLeft: TPanel;
    statMainInfo: TLMDStatusBar;
    dlgAbout: TLMDAboutDlg;
    VersInfo: TVersInfo;
    pnlBottom: TPanel;
    SplitterBottom: TSplitter;
    HotImages: TImageList;
    actionMain: TActionManager;
    actOpen: TFileOpen;
    actExit: TFileExit;
    menuMain: TMainMenu;
    menuFile: TMenuItem;
    menuHelp: TMenuItem;
    Open1: TMenuItem;
    Exit1: TMenuItem;
    N1: TMenuItem;
    N2: TMenuItem;
    actAbout: TAction;
    About1: TMenuItem;
    statMainInfoMessage2: TLMDSimpleLabel;
    statMainInfoMessage: TLMDSimpleLabel;
    edSectorNumHex: TEdit;
    edByteNumHex: TEdit;
    spinSector: TSpinButton;
    Label5: TLabel;
    Label6: TLabel;
    edByteNum: TNumberEdit;
    edSectorNum: TNumberEdit;
    actDisplayText: TAction;
    DisplayText1: TMenuItem;
    actParseBASIC: TAction;
    DisplayBASICListing1: TMenuItem;
    View1: TMenuItem;
    SelectFont1: TMenuItem;
    CoolBar1: TCoolBar;
    ActionToolBar1: TActionToolBar;
    ActionToolBar2: TActionToolBar;
    spinFontSize: TRxSpinEdit;
    cbFonts: TFontComboBox;
    DISFIXWarn: TProWarningDialog;
    actionFormatted: TActionManager;
    actSaveasText: TFileSaveAs;
    actPrintText: TPrintDlg;
    RichEditBold: TRichEditBold;
    RichEditItalic: TRichEditItalic;
    RichEditUnderline: TRichEditUnderline;
    RichEditStrikeOut: TRichEditStrikeOut;
    RichEditBullets: TRichEditBullets;
    RichEditAlignLeft: TRichEditAlignLeft;
    RichEditAlignRight: TRichEditAlignRight;
    RichEditAlignCenter: TRichEditAlignCenter;
    EditCut: TEditCut;
    EditCopy: TEditCopy;
    EditPaste: TEditPaste;
    EditSelectAll: TEditSelectAll;
    EditUndo: TEditUndo;
    EditDelete: TEditDelete;
    SearchFind: TSearchFind;
    SearchFindNext: TSearchFindNext;
    SearchReplace: TSearchReplace;
    SearchFindFirst: TSearchFindFirst;
    FontEdit: TFontEdit;
    actListNarrow: TAction;
    actFormatCompanion: TAction;
    DisplaycompanionText1: TMenuItem;
    Label8: TLabel;
    edContent: TNumberEdit;
    edContentchar: TEdit;
    actSaveasRTF: TAction;
    SaveDialog: TSaveDialog;
    actOpenImageFile: TAction;
    OpenImage: TOpenDialog;
    ActionFileList: TActionManager;
    actDisplayfile: TAction;
    actExtractFile: TAction;
    statSecondary: TLMDStatusBar;
    statSecondaryInfoMessage2: TLMDSimpleLabel;
    statSecondaryInfoMessage1: TLMDSimpleLabel;
    actCatalog: TAction;
    OpenDisketteImage1: TMenuItem;
    N3: TMenuItem;
    HelpContents1: THelpContents;
    Contents1: TMenuItem;
    FilePrintSetup: TFilePrintSetup;
    PagesTI: TPageControl;
    tabImageInfo: TTabSheet;
    pnlDisplays: TPanel;
    lblSectorCount: TLabel;
    lblSectorsPerTrack: TLabel;
    lblTracksPerSide: TLabel;
    lblSides: TLabel;
    lblDensity: TLabel;
    Label9: TLabel;
    Label7: TLabel;
    Label10: TLabel;
    edDiskName: TEdit;
    edSectorCount: TNumberEdit;
    edSectorsPerTrack: TNumberEdit;
    edTracksPerSide: TNumberEdit;
    edSides: TNumberEdit;
    edDensity: TNumberEdit;
    ckProtected: TCheckBox;
    gridDiskBitMap: TDrawGrid;
    Splitter1: TSplitter;
    ActionToolBar8: TActionToolBar;
    lbCatalog: TLMDListBox;
    tabFileInfo: TTabSheet;
    pnlTopRight: TPanel;
    lblTIFileType: TLabel;
    Label4: TLabel;
    lblSectors: TLabel;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    lblTIFileName: TLabel;
    gridHeader: TStringGrid;
    rgProgram: TRadioGroup;
    rgProtected: TRadioGroup;
    edTISectorCountHex: TEdit;
    edEOFOffsetHex: TEdit;
    rgDataType: TRadioGroup;
    rgDataSize: TRadioGroup;
    edRecSizeHex: TEdit;
    edRecCountHex: TEdit;
    edRecPerSectorHex: TEdit;
    edTISectorCount: TNumberEdit;
    edEOFOffset: TNumberEdit;
    edRecSize: TNumberEdit;
    edRecCount: TNumberEdit;
    edRecPerSector: TNumberEdit;
    ckTIWriter: TCheckBox;
    ckTIFILESHeader: TCheckBox;
    ckBASIC: TCheckBox;
    ckCompanion: TCheckBox;
    tabRawSectors: TTabSheet;
    tabRawText: TTabSheet;
    tabFormatted: TTabSheet;
    gridSectors: TAdvColumnGrid;
    memRawText: TMemo;
    CoolBar2: TCoolBar;
    ActionToolBar4: TActionToolBar;
    ActionToolBar5: TActionToolBar;
    ActionToolBar6: TActionToolBar;
    ActionToolBar7: TActionToolBar;
    pnTopRight: TPanel;
    splitVert: TSplitter;
    DriveComboBox1: TDriveComboBox;
    DirectoryListBox1: TDirectoryListBox;
    FileListBox1: TFileListBox;
    memFormatted: TRxRichEdit;
    procedure statMainVariableClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure actAboutExecute(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure gridSectorsOriginalClick(Sender: TObject);
    procedure actOpenAccept(Sender: TObject);
    procedure actionMainUpdate(Action: TBasicAction; var Handled: Boolean);
    procedure ReadTIFilesSectors(Sender: TObject);
    procedure spinSectorUpClick(Sender: TObject);
    procedure spinSectorDownClick(Sender: TObject);
    procedure DisplayTISector(SecNum: integer; Sec: array of char);
    procedure actDisplayTextExecute(Sender: TObject);
    procedure actParseBASICExecute(Sender: TObject);
    procedure NewFonts(Sender: TObject);
    procedure actPrintTextAccept(Sender: TObject);
    procedure actSaveAsTextAccept(Sender: TObject);
    procedure dlgAboutCustomClick(Sender: TObject);
    procedure FontEditAccept(Sender: TObject);
    procedure actListNarrowExecute(Sender: TObject);
    procedure ckBasicClick(Sender: TObject);
    procedure ckCompanionClick(Sender: TObject);
    procedure ckTIWriterClick(Sender: TObject);
    procedure actionFormattedUpdate(Action: TBasicAction;
      var Handled: Boolean);
    procedure actFormatCompanionExecute(Sender: TObject);
    procedure actSaveasRTFExecute(Sender: TObject);
    procedure actOpenImageFileExecute(Sender: TObject);
    procedure edSectorNumExit(Sender: TObject);
    procedure ClearTIFILESFileInfo;
    procedure ClearDisketteInfo;
    procedure ClearCurrentDiskette;
    procedure ClearMemos;
    procedure ClearCurrentFile;
    procedure actDisplayfileExecute(Sender: TObject);
    procedure actExtractFileExecute(Sender: TObject);
    procedure actCatalogExecute(Sender: TObject);
    procedure MakeTIFilesHeader(var ThisFile: TTIFile);
    procedure gridDiskBitMapDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure lbCatalogDblClick(Sender: TObject);
    procedure actOpenBeforeExecute(Sender: TObject);
    procedure FileListBox1DblClick(Sender: TObject);
  private
    BuildString: string;
    { Private declarations }
  public
    procedure DisplayHint(Sender: TObject);
    { Public declarations }
  end;

var
  frmMain: TfrmMain;
  TIHeaderBuf: array[1..16] of Char;
  TIFullHeaderBuf: array[1..128] of Char;
  TISectorBuf: array[0..255] of Char;
  //
  CurrentSector: integer;
  MaxSectors: integer;
  //
  CurrentFile: TTIFile;          // one file at a time
  CurrentDisk: TTIDisk;          // one disk at a time
  //
  DisplayingFile: boolean;
  DisplayingDiskImage: boolean;
  DisplayingDiskFile: boolean;
  ResizingGrid: boolean;
  //
  LineNumStart: integer;
  LineNumEnd: integer;
  ProgramTableEnd: integer;
  ProgramTableStart: integer;
  //
  PrintMargins: TRect;

implementation

{$R *.DFM}

procedure TfrmMain.statMainVariableClick(Sender: TObject);
begin
  statMainVariable.Visible := False;
  case Ord(statMainVariable.Info) of
  0: statMainVariable.Info := Succ(statMainVariable.Info);
  1: statMainVariable.Info := Succ(statMainVariable.Info);
  2..14:  while (Ord(statMainVariable.Info) <> 15) do
    statMainVariable.Info := Succ(statMainVariable.Info);
  15..24: while (Ord(statMainVariable.Info) <> 25) do
    statMainVariable.Info := Succ(statMainVariable.Info);
  25..39: while (Ord(statMainVariable.Info) <> 40) do
    statMainVariable.Info := Succ(statMainVariable.Info);
  40: while (Ord(statMainVariable.Info) <> 0) do
    statMainVariable.Info := Pred(statMainVariable.Info);
  end;
  if Ord(statMainVariable.Info) = 40 then
        statMainVariable.UseStandardCaption := false
  else
        statMainVariable.UseStandardCaption := true;
  statMainVariable.Visible := True;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
var found: integer;
  exerec: TSearchRec;
begin
  found:=findfirst(application.exename,faAnyFile,exerec);
  versinfo.FileName := exerec.Name;
  Buildstring :='Build ' + versinfo.ModuleVersionNumber
   + ': ' + FormatDateTime('dddd, mmmm d, yyyy '
   + '"at" h:mm am/pm', filedatetodatetime(exerec.Time));
  FindClose(exerec);
  //
  statMainInfoMessage.Caption := BuildString;
  statMainInfoMessage.font.Color := clNavy;
  statMainInfoMessage2.Caption := '';
  dlgAbout.CaptionTitle := Buildstring;
  dlgAbout.Appname := versinfo.ProductName;
  dlgAbout.Copyright := versinfo.Copyright;
  dlgAbout.Description := versinfo.Comments;
  dlgAbout.Version := versinfo.ModuleVersionNumber;
  dlgAbout.EMail := 'rgm@fabbnet.net';
  dlgAbout.Jump := 'http://www.fabbnet.net/delphi.htm';
  //
  Application.OnHint := DisplayHint; Application.OnHint := DisplayHint;
end;

procedure TfrmMain.DisplayHint(Sender: TObject);
begin
  statMainHint.Caption := GetLongHint(Application.Hint);
end;

procedure TfrmMain.actAboutExecute(Sender: TObject);
begin
  dlgAbout.CaptionTitle :=
    Buildstring;
  dlgAbout.ExecuteEnh(frmMain);
end;

// ***********************************************
// * Application-Specific Procedure Code         *
// ***********************************************

//
//  Useful Internal Functions
//

//
//  Some 'borrowed' conversion functions
//

{function IntToHEX2(Value: Integer): String;
var B: Byte;
    I: Integer;
begin
   Result := '';
   for I := 8 downto 1 do
     begin
        B := Byte(Value) and $F;
        Value := Value shr 4;
        if B > 9 then B := B-10+$41 else B := B+$30;
        if Result = '' then Result := Char(B) else Insert(Char(B), Result, 1);
     end;
end;

function SToHEX(S: String): String;
//function to convert string to sequence of Hex Values
//  taken from original DIO code
var B1, B2: Byte;
    I: Integer;
begin
   Result := '';
   if Length(S) = 0 then Exit;
   for I := 1 to Length(S) do
       begin
          B1 := Byte(S[I]);
          B2 := B1 and $F;
          B1 := (B1 shr 4) and $F;
          if B2 > 9 then B2 := B2-10+$41 else B2 := B2+$30;
          if B1 > 9 then B1 := B1-10+$41 else B1 := B1+$30;
          Result := Result+Char(B1)+Char(B2)+' ';
       end;
end;}

procedure SetRichEdMargins(ARichEdit: TRxRichEdit; Arect: TRect);
//
// routine to set Richedit pixel margins based on a
//    page setup 1/1000" margins Rect passed as parameter
//

var
   NewRect: TRect;
   LogX, LogY: integer;
begin
   { Margins are in 1000ths of an inch from each edge of the paper. }
   { PageRect is the outline of the printable area in pixels.       }
   {                                                                }
   { The docs say twips, but the source code for TRichEdit98.Print    }
   { converts pixels...                                             }
   {                                                                }
   { We still have to convert the Right and Bottom values           }
   { since they are the distance from the edge of their respective  }
   { sides of the paper, not the distance from the other margin     }
   LogX := GetDeviceCaps(Printer.Handle, LOGPIXELSX);
   LogY := GetDeviceCaps(Printer.Handle, LOGPIXELSY);
   //
   NewRect.Left   := (ARect.Left * LogX div 1000) -
      GetDeviceCaps(Printer.Handle, PHYSICALOFFSETX);
   NewRect.Top    := ARect.Top * LogY div 1000 -
      GetDeviceCaps(Printer.Handle, PHYSICALOFFSETY);
   //
   NewRect.Right  := Printer.PageWidth - (ARect.Right * LogX div 1000) +
      GetDeviceCaps(Printer.Handle, PHYSICALOFFSETX);
   NewRect.Bottom := Printer.PageHeight - (ARect.Bottom * LogY div 1000) +
      GetDeviceCaps(Printer.Handle, PHYSICALOFFSETY);
   //
   ARichEdit.PageRect:= NewRect;
end;

{procedure SetRichEdNarrowMargins(ARichEdit: TRichEdit98; Arect: TRect;
             AWidth: integer);
var
   NewRect: TRect;
   LogX, LogY, WidthInPixels: integer;
begin
   WidthInPixels := printer.Canvas.TextWidth(StringofChar(#32, AWidth));
   //
   LogX := GetDeviceCaps(Printer.Handle, LOGPIXELSX);
   LogY := GetDeviceCaps(Printer.Handle, LOGPIXELSY);
   //
   NewRect.Left   := (ARect.Left * LogX div 1000) -
      GetDeviceCaps(Printer.Handle, PHYSICALOFFSETX);
   NewRect.Top    := ARect.Top * LogY div 1000 -
      GetDeviceCaps(Printer.Handle, PHYSICALOFFSETY);
   //
   NewRect.Right := NewRect.Left + (trunc(WidthInPixels div LogX) * 1000);
   //NewRect.Bottom := Printer.PageHeight - (ARect.Bottom * LogY div 1000) +
      GetDeviceCaps(Printer.Handle, PHYSICALOFFSETY);
   //
   ARichEdit.PageRect:= NewRect;
end;}

function DecodeType(T, RSize: integer): string;
 //
 //  decode 'status flag' byte into a tight little 7-byte string
 //
 begin
   case T of
   // Program/Image
   1: Result :=
         format('IMAGE U', [RSize]); //image u  0000 0001  = 1
   9: Result :=
         format('IMAGE P', [RSize]); //image p  0000 1001  = 5
   // DIS/VAR
   0: Result :=
         format('DF%3d U', [RSize]); //dv sz u  0000 0000  = 0
   4: Result :=
         format('DF%3d P', [RSize]); //dv sz p  0000 1000  = 8
   // INT/VAR
   2: Result :=
         format('IF%3d U', [RSize]); //iv sz u  0000 0010  = 2
   10: Result :=
         format('IF%3d P', [RSize]); //iv sz u  0000 1010  = 10
   // DIS/FIX
   128: Result :=
         format('DV%3d U', [RSize]); //dv sz p  1000 0000  = 128
   136: Result :=
         format('DV%3d P', [RSize]); //dv sz p  1000 1000  = 136
   // INT/FIX
   130: Result :=
         format('IV%3d U', [RSize]); //iv sz u  1000 0010  = 130
   138: Result :=
         format('IV%3d P', [RSize]); //iv sz u  1000 1010  = 138
   end;
end;

function CheckforComp: boolean;
begin
//*************************************
//*                                   *
//*  companion was my original        *
//*  favorite word processor. It      *
//*  has a unique file format.        *
//*                                   *
//*************************************
  Result := False;
  if ((ord(CurrentFile.SectorList[0].contents[0]) =
       ord(CurrentFile.SectorList[0].contents[1]) + 1)
     and
      (Copy(CurrentFile.Header.FileType, 1, 5) = 'IV254'))
  then Result := True;
end;

function CheckforWriter: boolean;
begin
//*************************************
//*                                   *
//*  TI Writer files are just DV 80   *
//*  so for now we just say TRUE here *
//*  and let the user override with   *
//*  the GUI                          *
//*                                   *
//*************************************
  Result := Copy(CurrentFile.Header.FileType, 1, 5) = 'DV 80';
end;

function CheckforBasic: boolean;
var word1: integer;
begin
//*************************************
//*                                   *
//*  the first word of a basic file   *
//*  is equal to the XOR of the       *
//*  second and third words. Logical? *
//*                                   *
//*  also capture here the critical   *
//*  line and statement table         *
//*  addresses from the BASIC header  *
//*                                   *
//*************************************
  word1 :=
    (256 * ord(CurrentFile.SectorList[0].contents[0]))
      + ord(CurrentFile.SectorList[0].contents[1]);       //test word
  LineNumEnd :=
    (256 * ord(CurrentFile.SectorList[0].contents[2]))
      + ord(CurrentFile.SectorList[0].contents[3]);       //End of Line
  //                                                      //Number Table
  ProgramTableStart := LineNumEnd + 2;
  //
  LineNumStart :=
    (256 * ord(CurrentFile.SectorList[0].contents[4]))
      + ord(CurrentFile.SectorList[0].contents[5]);       //Start of Line
  ProgramTableEnd :=                                      //Number Table
    (256 * ord(CurrentFile.SectorList[0].contents[6]))
      + ord(CurrentFile.SectorList[0].contents[7]);       //"Higest Address"
  //                                                      // of statement table
  Result := (word1 = (LineNumEnd xor LineNumStart));      // <-- the test
end;

function UsedSegment(SegNumber: integer): boolean;
//***********************************************
//*                                             *
//*  "turn on" bit (grid square) for used       *
//*  segments. Assumes an existing and valid    *
//*  CurrentDisk.Info.Bitmap                    *
//*                                             *
//***********************************************
var BitMapIndex, BitIndex: integer;
begin
Result := False;
//
// Don't bother unless we have a disk image in memory
//
if DisplayingDiskImage then begin
  //
  // which word?
  //
  BitMapIndex := SegNumber div 8;
  BitIndex := SegNumber mod 8;
  //
  // AND with correct bit
  //
  case bitIndex of
    0: if (1 and ord(CurrentDisk.Info.Bitmap[BitMapIndex])) = 1 then
               Result := true;
    1: if (2 and ord(CurrentDisk.Info.Bitmap[BitMapIndex])) = 2 then
               Result := true;
    2: if (4 and ord(CurrentDisk.Info.Bitmap[BitMapIndex])) = 4 then
               Result := true;
    3: if (8 and ord(CurrentDisk.Info.Bitmap[BitMapIndex])) = 8 then
               Result := true;
    4: if (16 and ord(CurrentDisk.Info.Bitmap[BitMapIndex])) = 16 then
               Result := true;
    5: if (32 and ord(CurrentDisk.Info.Bitmap[BitMapIndex])) = 32 then
               Result := true;
    6: if (64 and ord(CurrentDisk.Info.Bitmap[BitMapIndex])) = 64 then
               Result := true;
    7: if (128 and ord(CurrentDisk.Info.Bitmap[BitMapIndex])) = 128 then
               Result := true;
    end;
  end;
end;

function ThisToken(a: char): string;
//*****************************************
//*                                       *
//*     BASIC token interpreter           *
//*                                       *
//*****************************************
begin
case ord(a) of
    //
    // Basic Tokens
    //
    //
    //      these should never be parsed
    //
    0  : ThisToken := ''; //'end of line marker';
    255: ThisToken := ''; //'end of file marker';
    //
    //      these are special cases
    //
    199: ThisToken := '"'; //'quoted string';   //>C7  followed by length byte
                                                //     trailing double quote not
                                                //     and the actual string.
                                                //     the trailing doublequote
                                                //     is only implied, and must
                                                //     be supplied by the
                                                //     decoding routine

    200: ThisToken := '';  //'string';          //>C8  followed by length byte
    201: ThisToken := '';  //'Stmt #';          //>C9  followed by two byte int
    //
    //   a sort of special case?
    //
    253: ThisToken := '#'; //(files, a.k.a. )   //>FD  nothing special, really,
                           // "unquoted string  //     but it's followed by the
                           // NUMERIC"          //     >C8 token and the file
                           // by Millers        //     number as a literal
                           // Graphics          //     string!!
    //
    //  the rest of these are just parsed, one for one
    //    positioning/supplying spaces can be a problem
    //
    129: ThisToken := ' ELSE ';
    132: ThisToken := ' IF ';
    133: ThisToken := ' GO';
    134: ThisToken := ' GOTO';
    135: ThisToken := ' GOSUB';
    136: ThisToken := ' RETURN ';
    137: ThisToken := ' DEF ';
    138: ThisToken := ' DIM ';
    139: ThisToken := ' END ';
    140: ThisToken := ' FOR ';
    141: ThisToken := ' LET ';
    142: ThisToken := ' BREAK ';
    143: ThisToken := ' UNBREAK ';
    144: ThisToken := ' TRACE ';
    145: ThisToken := ' UNTRACE ';
    146: ThisToken := ' INPUT ';
    147: ThisToken := ' DATA ';
    148: ThisToken := ' RESTORE ';
    149: ThisToken := ' RANDOMIZE ';
    150: ThisToken := ' NEXT ';
    151: ThisToken := ' READ ';
    152: ThisToken := ' STOP ';
    153: ThisToken := ' DELETE ';
    154: ThisToken := ' REM ';
    155: ThisToken := ' ON ';
    156: ThisToken := ' PRINT ';
    157: ThisToken := ' CALL ';
    158: ThisToken := ' OPTION ';
    159: ThisToken := ' OPEN ';
    160: ThisToken := ' CLOSE ';
    161: ThisToken := ' SUB ';
    162: ThisToken := ' DISPLAY ';
    164: ThisToken := ' ACCEPT ';
    169: ThisToken := ' RUN ';
    176: ThisToken := ' THEN ';
    177: ThisToken := ' TO ';
    178: ThisToken := ' STEP ';
    179: ThisToken := ','; //(comma)
    180: ThisToken := ';'; //(semi)
    181: ThisToken := ':'; //(colon)
    182: ThisToken := ') ';
    183: ThisToken := '(';
    184: ThisToken := ' & ';
    190: ThisToken := ' = ';
    191: ThisToken := ' < ';
    192: ThisToken := ' > ';
    193: ThisToken := ' + ';
    194: ThisToken := ' -';   // no space for negative numbers
    195: ThisToken := ' * ';
    196: ThisToken := '/';
    197: ThisToken := '^';
    202: ThisToken := ' EOF';
    203: ThisToken := ' ABS';
    204: ThisToken := ' ATN';
    205: ThisToken := ' COS';
    206: ThisToken := ' EXP';
    207: ThisToken := ' INT';
    208: ThisToken := ' LOG';
    209: ThisToken := ' SGN';
    210: ThisToken := ' SIN';
    211: ThisToken := ' SQR';
    212: ThisToken := ' TAN';
    213: ThisToken := ' LEN';
    214: ThisToken := ' CHR$';
    215: ThisToken := ' RND';
    216: ThisToken := ' SEG$';
    217: ThisToken := ' POS';
    218: ThisToken := ' VAL';
    219: ThisToken := ' STR$';
    220: ThisToken := ' ASC';
    222: ThisToken := ' REC';
    241: ThisToken := ' BASE ';
    243: ThisToken := ' VARIABLE ';
    244: ThisToken := ' RELATIVE ';
    245: ThisToken := ' INTERNAL ';
    246: ThisToken := ' SEQUENTIAL ';
    247: ThisToken := ' OUTPUT ';
    248: ThisToken := ' UPDATE ';
    249: ThisToken := ' APPEND ';
    250: ThisToken := ' FIXED ';
    251: ThisToken := ' PERMANENT ';
    252: ThisToken := ' TAB ';
    //
    //  XBasic only
    //
    130: ThisToken := ' :: ';
    131: ThisToken := '!';
    163: ThisToken := ' IMAGE ';
    165: ThisToken := ' ERROR ';
    166: ThisToken := ' WARNING ';
    167: ThisToken := ' SUBEXIT ';
    168: ThisToken := ' SUBEND ';
    170: ThisToken := ' LINPUT ';
    186: ThisToken := ' OR ';
    187: ThisToken := ' AND ';
    188: ThisToken := ' XOR ';
    189: ThisToken := ' NOT ';
    221: ThisToken := ' PI ';
    223: ThisToken := ' MAX';
    224: ThisToken := ' MIN';
    225: ThisToken := ' RPT$';
    232: ThisToken := ' NUMERIC ';
    233: ThisToken := ' DIGIT ';
    234: ThisToken := ' UALPHA ';
    235: ThisToken := ' SIZE ';
    236: ThisToken := ' ALL ';
    237: ThisToken := ' USING ';
    238: ThisToken := ' BEEP ';
    239: ThisToken := ' ERASE ';
    240: ThisToken := ' AT';
    254: ThisToken := ' VALIDATE ';
    //
    //  everything else
    //
    else ThisToken := chr(ord(a));
    //
    end;  //case
end;

function ReadTIImage(AFilename: string): boolean;
  //***************************************
  //*  subprocedure to decode nybbles     *
  //*  in a 3-byte cluster record         *
  //***************************************
  procedure Nybbler(var X: TTISegmentGroup);
  begin
     X.StartSeg := strtoint('$' +
      (copy((inttohex(ord(X.cluster[1]), 2)), 2, 1)
         + inttohex(ord(X.cluster[0]), 2)));
     //
     X.EndOffset := strtoint('$' +
      (inttohex(ord(X.cluster[2]), 2)
         + copy((inttohex(ord(X.cluster[1]), 2)), 1, 1)));
  end;
//
//   main function code
//
var FromF: file;
    SS: string[10];
    NumRead, i, j, k: Integer;
    FDRList: TStrings;
    ClusterPointer: integer;
begin
  Result := False;
  frmMain.statMainInfoMessage.Caption := '';
  //
  //  Open File and Read Sector 0;
  //
  FDRList := TStringList.Create;
  AssignFile(FromF, AFileName);
  Reset(FromF, 1);
  //
  try
    ResizingGrid := true;            // set flag to kill OnDrawCell
    BlockRead(FromF, TISectorBuf, SizeOf(TISectorBuf), NumRead);
    //
    SS := '';
    for i := 0 to 9 do
      SS := SS +  TISectorBuf[i];
    CurrentDisk.Info.Diskname := SS;
    SS := '';
    CurrentDisk.Info.SectorCount := (256 * ord(TISectorBuf[10]))
                                    +  ord(TISectorBuf[11]);
    CurrentDisk.Info.SectorsPerTrack := ord(TISectorBuf[12]);
    //
    //case CurrentDisk.Info.SectorCount of
    //  360: begin
    //       frmMain.gridDiskBitMap.ColCount := 8;
    //       frmMain.gridDiskBitMap.Width := 20;
    //       frmMain.gridDiskBitMap.Height := 94;
    //       SetLength(CurrentDisk.Info.Bitmap, 45);
    //       end;
    //  720: begin
    //       frmMain.gridDiskBitMap.ColCount := 16;
    //       frmMain.gridDiskBitMap.Width := 36;
    //       frmMain.gridDiskBitMap.Height := 94;
    //       SetLength(CurrentDisk.Info.Bitmap, 90);
    //       end;
    //  1440: begin
    //        frmMain.gridDiskBitMap.ColCount := 32;
    //        frmMain.gridDiskBitMap.Width := 68;
    //        frmMain.gridDiskBitMap.Height := 94;
    //        SetLength(CurrentDisk.Info.Bitmap, 180);
    //        end;
     // end; //case

    frmMain.gridDiskBitMap.ColCount :=
      CurrentDisk.Info.SectorCount div 45;
    frmMain.gridDiskBitMap.Width :=
     (2 * frmMain.gridDiskBitMap.ColCount) + 4;
    frmMain.gridDiskBitMap.Height :=
     (2 * frmMain.gridDiskBitMap.RowCount) + 4;
    //frmMain.gridDiskBitMap.DefaultColWidth :=
    //  2 * (1440 div CurrentDisk.Info.SectorCount);
    //
    SetLength(CurrentDisk.Info.Bitmap,
      CurrentDisk.Info.SectorCount div 8);
    //
    for i := 13 to 15 do
      SS := SS +  TISectorBuf[i];
    CurrentDisk.Info.DSRMark := SS;
    //
    if TISectorBuf[16] = 'P' then
      CurrentDisk.Info.diskProtected := true
    else CurrentDisk.Info.diskProtected := False;
    //
    CurrentDisk.Info.TracksPerSide := ord(TISectorBuf[17]);
    CurrentDisk.Info.Sides := ord(TISectorBuf[18]);
    CurrentDisk.Info.Density := ord(TISectorBuf[19]);
    //
    // Read Bitmap
    //
    for i := 56 to (56 + (CurrentDisk.Info.SectorCount div 8)) do
      CurrentDisk.Info.Bitmap[i - 56] :=  TISectorBuf[i];
    //
    //  kinda false test, but it SHOULD be true
    //
    if CurrentDisk.Info.DSRMark = 'DSK' then begin
    //
      Result := True;
      DisplayingDiskImage := True;
      DisplayingDiskFile := False;
      DisplayingFile := False;
      MaxSectors := CurrentDisk.Info.SectorCount - 1;
      //
      //  Read Sector 1 list of FDR addresses and Set File Count
      //
      BlockRead(FromF, TISectorBuf, SizeOf(TISectorBuf), NumRead);
      i := 0;
      while (256 * ord(TISectorBuf[i]) + ord(TISectorBuf[i + 1])) <> 0 do begin
        FDRList.Add(inttostr((256 * ord(TISectorBuf[i])
          + ord(TISectorBuf[i + 1]))));
        inc(i); inc(i);                   // skip to next byte pair
        end;
      CurrentDisk.Info.FileCount := FDRList.Count;
      //
      //  adjust the array sizes
      //
      SetLength(CurrentDisk.SectorList, CurrentDisk.Info.SectorCount);
      SetLength(CurrentDisk.Directory, CurrentDisk.Info.FileCount);
      end;
    //
    // Read ALL Sectors (yes, read #0 and #1 again...)
    //
    if Result then begin
      Reset(FromF, 1);      // back to the future
      for i := 0 to CurrentDisk.Info.SectorCount - 1 do begin
        frmMain.statMainInfoMessage.Caption :=
          'Reading Sector ' + inttostr(i);
        BlockRead(
              FromF,
              CurrentDisk.SectorList[i].Contents,
              SizeOf(CurrentDisk.SectorList[i].Contents),
              NumRead);
        CurrentDisk.SectorList[i].EndOffset := NumRead;
        end;
        CurrentDisk.SectorsRead := true;
      end;
    //
    // Now we have all Sectors in Memory
    //   so we can build FDR array
    //
    for i := 0 to FDRList.Count - 1 do begin
      frmMain.statMainInfoMessage.Caption :=
          'FDR array entry ' + inttostr(i);
      with CurrentDisk.Directory[i] do begin
        Address := strtoint(FDRList[i]);
        SS := '';
        for j := 0 to 9 do
          SS := SS + CurrentDisk.SectorList[Address].Contents[j];
        //
        Filename := SS;
        // byte 12
        StatusFlag := ord(CurrentDisk.SectorList[Address].Contents[12]);
        // byte 13
        RecordsPerSector :=
           ord(CurrentDisk.SectorList[Address].Contents[13]);
        // bytes 14-15
        SectorCount :=
           (256 * ord(CurrentDisk.SectorList[Address].Contents[14]))
            + ord(CurrentDisk.SectorList[Address].Contents[15]);
        // byte 16
        LastByteInLastSector :=
           ord(CurrentDisk.SectorList[Address].Contents[16]);
        // byte 17
        RecordLength := ord(CurrentDisk.SectorList[Address].Contents[17]);
        // bytes 18 and 19 but swapped!
        NumberRecordsSectors :=
           (256 * ord(CurrentDisk.SectorList[Address].Contents[19]))
            + ord(CurrentDisk.SectorList[Address].Contents[18]);
        //frmMain.memDebug.Lines.Add(CurrentDisk.Directory[i].Filename);
        //
        clusterpointer := 28;
        //
        for k := 0 to 74 do begin
          Segmentlist[k].Cluster[0]
            := CurrentDisk.SectorList[Address].Contents[clusterpointer];
          Segmentlist[k].Cluster[1]
            := CurrentDisk.SectorList[Address].Contents[clusterpointer + 1];
          Segmentlist[k].Cluster[2]
            := CurrentDisk.SectorList[Address].Contents[clusterpointer + 2];
          //  and decode it
          Nybbler(SegmentList[k]);
          clusterpointer := clusterpointer + 3;
          end;
        end;
      end;
    //
  finally
    CloseFile(FromF);
    FDRList.Free;                    //<-- no memory leak
    ResizingGrid := False;           // Turn OnDrawCell back on
  end;
end;

function ReadTIFilesHeader(AFilename: String;
  var AHeader: TTIFilesHeader): boolean;
 { The oft-quoted FAQ I below is wrong: it reverses the VAR/FIX bit,
   or maybe I'm just dyslexic, but cut and paste can't lie...
   or maybe Telco is mis-building the Header? doubt it...

 "Offset  8 -- # sectors multiply by 256 to get actual bytes
  Offset 10 -- Status flags:
  Bit#  Description
   0    0 = Data file
        1 = Program (memory image) file       0000 0001
   1    0 = Display                           0000 0010
        1 = Internal
   2        Reserved (0)
   3    0 = Unprotected                       0000 1000
        1 = Protected
  4-6       Reserved (0)
   7    0 = Variable-length recs              1000 0000
        1 = Fixed-length records              // this is the backwards part
                                              // 1 = VAR; 0 = FIX
  "Offset 11 -- # recs per sector
               Divide 256 by logical record size.
               Only valid for FIXED LENGTH records.
               I'm pretty sure the field is meaningless
               for VARIABLE LENGTH records.
               A DF80 file has 3 records per sector...
               an IF128 has 2.  DV80 will probably tell you 3,
               but since its sectors may have more than 3 records,
               it's meaningless.
  "Offset 12 -- EOF offset
               You're quite correct... number of bytes used
               in last sector. In DV80 files, points to
               an >FF if I recall correctly.
  "Offset 13 -- Logical record size
               Well...  maximum number of chars per record...
               thus 80 in DV80, 163 in DV163, etc...
               Note that for variable-length records,
               the actual record length can be up to n+1
               (1byte for size storage)
  "Offset 14 -- # of lev 3 records = total # of records allocated
               You went and found an obtuse definition.
               Corresponds to offset 8 in extra params
               block of TDIF/TDOF. Simply put,
               how many records are in this file???
  "Offset 16 -- no clue! I'll have to check my old sources for this.}

  procedure DecodeStatusByte;
  //****************************************************
  //*  This uses StatusFlag for some extra stuff       *
  //*     that I define separately in header record    *
  //****************************************************
  begin
   AHeader.ProgType := (AHeader.StatusFlag and 1) = 1;
   //
   if (AHeader.StatusFlag and 2) = 2 then
     AHeader.DataType := 'INT'
   else
     AHeader.DataType := 'DIS';
   //
   AHeader.ProtectBit := (AHeader.StatusFlag and 8) = 8;
   //
   if (AHeader.StatusFlag and 128) = 128 then
     AHeader.DataSize := 'VAR'
   else
     AHeader.DataSize := 'FIX';
   //*********************************************************
   //*  Includes this call to the external DeCodeType        *
   //*  routine, which is separated to be available to the   *
   //*  disk image catalog routines. DeCodeType is basically *
   //*  a sledgehammer approach since there are only a few   *
   //*  possible combinations                                *
   //*********************************************************
   AHeader.FileType := DecodeType(AHeader.StatusFlag, AHeader.RecSize);
  end;
//
//  Main Code Block
//
var
  i, j: integer;
  FromF: file;
  SS: string;
  NumRead: Integer;
begin
  Result := False;
  SS := 'NOT  TI';                           // prove me wrong
  Move(SS[1], TIHeaderBuf[1], Length(SS));
  //
  AssignFile(FromF, AFileName);
  Reset(FromF, 1);
  //
  try
    BlockRead(FromF, TIHeaderBuf, SizeOf(TIHeaderBuf), NumRead);
  finally
    CloseFile(FromF);
  end;
  //
  Move(TIHeaderBuf[1], AHeader.HeaderString, Sizeof(AHeader.HeaderString));
  Result := (Aheader.HeaderString = 'TIFILES');
  //
  if Result = true then begin
    DisplayingFile:= True;
    AHeader.Sectors := (256 * ord(TIHeaderBuf[9])) + ord(TIHeaderBuf[10]);
    MaxSectors := AHeader.Sectors - 1;
    AHeader.StatusFlag := ord(TIHeaderBuf[11]);
    AHeader.RecordsPerSec := ord(TIHeaderBuf[12]);
    AHeader.EOFOffset := ord(TIHeaderBuf[13]);
    AHeader.RecSize := ord(TIHeaderBuf[14]);
    //
    //  The 'record count' is only for
    //   variable length records but we always read it
    //
    AHeader.RecCount := (256 * ord(TIHeaderBuf[16])) + ord(TIHeaderBuf[15]);
    //
    DecodeStatusByte;
    //
    //  Display the actual header 'Header' (1st 16 bytes) in hex
    //
    for i := 0 to 1 do
      for j := 0 to 7 do
       frmMain.gridHeader.Cells[j, i] :=
         format(   '%.2x', [  ord(TIHeaderBuf[(i * 8) + j + 1])  ]   );
    //
    end;
end;

//***************************************
//*  Published procedures - subs which  *
//*  are part of the form object        *
//***************************************

procedure TfrmMain.MakeTIFilesHeader(var ThisFile: TTIFile);
//***************************************
//*  This is separated so it can be     *
//*  called for a stand-alone file or   *
//*  one extracted from a DOAD file.    *
//***************************************
begin
  ThisFile.Header.HeaderString := 'TIFILES';
  ThisFile.Header.Sectors :=
    CurrentDisk.Directory[lbCatalog.ItemIndex].SectorCount;
  ThisFile.Header.StatusFlag :=
    CurrentDisk.Directory[lbCatalog.ItemIndex].StatusFlag;
  ThisFile.Header.RecSize :=
    CurrentDisk.Directory[lbCatalog.ItemIndex].Recordlength;
  //
  Thisfile.Header.ProgType := (Thisfile.Header.StatusFlag and 1) = 1;
  //
  if (Thisfile.Header.StatusFlag and 2) = 2 then
   Thisfile.Header.DataType := 'INT'
  else
   Thisfile.Header.DataType := 'DIS';
  //
  Thisfile.Header.ProtectBit := (Thisfile.Header.StatusFlag and 8) = 8;
  //
  if (Thisfile.Header.StatusFlag and 128) = 128 then
   Thisfile.Header.DataSize := 'VAR'
  else
   Thisfile.Header.DataSize := 'FIX';
  //
  Thisfile.Header.FileType :=
    DecodeType(Thisfile.Header.StatusFlag, Thisfile.Header.RecSize);
  //
  ThisFile.Header.RecordsPerSec :=
    CurrentDisk.Directory[lbCatalog.ItemIndex].RecordsPerSector;
  ThisFile.Header.EOFOffset :=
    CurrentDisk.Directory[lbCatalog.ItemIndex].LastByteInLastSector;
  ThisFile.Header.RecCount :=
    CurrentDisk.Directory[lbCatalog.ItemIndex].NumberRecordsSectors;
  //
  MaxSectors := ThisFile.Header.Sectors - 1;
end;

//***************************************
//*  App-specific form-based routines   *
//***************************************

procedure TfrmMain.FormActivate(Sender: TObject);
var i: integer;
begin
  //
  // label gridSector cells
  //
  for i := 0 to gridSectors.RowCount - 1 do begin
      if ((i <= (gridSectors.ColCount div 2)) and (i <> 0)) then begin
        gridSectors.cells[i, 0] := format('%.2x', [i - 1]);
        gridSectors.cells[i + (gridSectors.ColCount div 2) , 0] :=
          format('%.2d', [i - 1]);
        end;
      if i <> 0 then gridSectors.cells[0, i] :=
        format('%.4x', [(i - 1) * ((gridSectors.colcount div 2))]);
      end;
  gridSectors.ColWidths[0] := 30;
  //
  CurrentFile.FormattedFlag := False;
  pagesTI.ActivePage := tabImageInfo; //opening screen
  //
  i := 0;
  while (i < cbFonts.Items.Count) and (cbFonts.Items[i] <> 'Lucida Sans Typewriter') do
    inc(i);
  //Result := i;
  cbFonts.itemindex := i;
  //NewFonts(Self);
  //cbFonts.ItemIndex := 0;
  NewFonts(Self);
  ClearDisketteInfo;
  ClearCurrentDiskette;
  ClearTIFilesFileInfo;
  ClearMemos;
  ResizingGrid := False;
  //
  DriveComboBox1.Drive := 'C';
  DirectoryListBox1.Directory := '\Execute\Burrsoft\Win994a\Disks\';;

end;

//
//  primary actions
//

procedure TfrmMain.actOpenAccept(Sender: TObject);
//***************************************
//*  Open stand-alone TI file           *
//***************************************
begin
  statMainInfoMessage.Caption := '';
  //pagesTI.ActivePage := tabFileInformation;
  //
  // de-allocate sector arrays and clean up generally
  //
  ClearCurrentFile;
  ClearTIFILESFileInfo;
  ClearDisketteInfo;
  ClearCurrentDiskette;
  ClearMemos;
  //
  //  Read and check file header
  //
  ckTIFilesHeader.Checked :=
   ReadTIFilesHeader(actOpen.Dialog.FileName, CurrentFile.Header);
  //
  //  Display header information
  //
  lblTIFileType.caption := CurrentFile.Header.FileType;
  rgProtected.Enabled := True;
  rgProgram.Enabled := True;
  //
  if CurrentFile.Header.ProtectBit then rgProtected.ItemIndex := 0
    else rgProtected.ItemIndex := 1;
  //
  if not CurrentFile.Header.ProgType then begin
    rgDataType.Visible := True;
    rgDataSize.Visible := True;
    label1.Visible := True;
    label3.Visible := True;
    edRecSizeHex.Visible := True;
    edRecSize.Visible := True;
    if CurrentFile.Header.DataSize = 'VAR' then begin
      label2.Visible := True;
      edRecCountHex.Visible := True;
      edRecCount.Visible := True;
      end;
    edRecPerSectorHex.Visible := True;
    edRecPerSector.Visible := True;
    rgProgram.ItemIndex := 1;
    if CurrentFile.Header.datasize = 'FIX' then
      rgDataSize.itemindex := 0 else rgDataSize.itemindex := 1;
    if CurrentFile.Header.dataType = 'DIS' then
      rgDataType.itemindex := 0 else rgDataType.itemindex := 1;
    edRecSizeHex.Text := inttohex(CurrentFile.Header.RecSize, 4);
    edRecCountHex.Text := inttohex(CurrentFile.Header.RecCount, 4);
    edRecPerSectorHex.Text := inttoHex(CurrentFile.Header.RecordsPerSec, 4);
    edRecSize.Value := CurrentFile.Header.RecSize;
    edRecCount.Value := CurrentFile.Header.RecCount;
    edRecPerSector.Value := CurrentFile.Header.RecordsPerSec;
    end
  else rgProgram.ItemIndex := 0;
  //
  edTISectorCountHex.Text := inttoHex(CurrentFile.Header.Sectors, 4);
  edEOFOffsetHex.Text := inttoHex(CurrentFile.Header.EOFOffset, 4);
  edTISectorCount.Value := CurrentFile.Header.Sectors;
  edEOFOffset.Value := CurrentFile.Header.EOFOffset;
  //
  //  Re-allocate Sector Array and
  //    Reset rest of CurrentFile Record
  //
  SetLength(CurrentFile.SectorList, CurrentFile.Header.Sectors);
  CurrentFile.SectorsRead := false;
  CurrentFile.PCName := actOpen.Dialog.FileName;
  lblTIFileName.caption := UpperCase(ExtractFileName(CurrentFile.PCName));
  //
  statMainInfoMessage2.Caption := CurrentFile.PCName;
  if not ckTIFILESHeader.Checked then
    statMainInfoMessage.Caption := '  Not a TI File?';
  //
  if CurrentFile.Header.HeaderString = 'TIFILES' then begin
    ReadTIFilesSectors(Self);
    statSecondaryInfoMessage1.Caption := 'TIFILES PC File in memory';
    CurrentFile.StandAlone := True;
    end;
end;

procedure TfrmMain.actOpenImageFileExecute(Sender: TObject);
//***************************************
//*   Open V9T9 DOAD disk image         *
//***************************************
var i, j: integer;
    SS: string;
begin
  OpenImage.Filter := 'WIN994A images|*.TIDisk|V9T9 images|*.V9|All Files|*.*';
  if OpenImage.Execute then begin
    screen.Cursor := crHourGlass;
    memFormatted.Clear;
    statMainInfoMessage.Caption := 'Stage One';
    //
    //  Reset all displays, records and arrays
    //
    statMainInfoMessage.Caption := 'Clearing File Info';
    ClearTIFilesFileInfo;
    statMainInfoMessage.Caption := 'Clearing Disk Display';
    ClearDisketteInfo;
    statMainInfoMessage.Caption := 'Clearing Disk Array';
    ClearCurrentDiskette;
    //
    //  Get valid disk info
    //
    statMainInfoMessage.Caption := 'Opening Image.......';
    if ReadTIImage(OpenImage.FileName) then begin
      edDiskName.text := CurrentDisk.Info.DiskName;
      ckProtected.Checked := CurrentDisk.Info.DiskProtected;
      edSectorCount.Value := CurrentDisk.Info.SectorCount;
      edSectorsPerTrack.Value := CurrentDisk.Info.SectorsPerTrack;
      edTracksPerSide.Value := CurrentDisk.Info.TracksPerSide;
      edSides.Value := CurrentDisk.Info.Sides;
      edDensity.Value := CurrentDisk.Info.Density;
      //
      // Display Directory
      //
      lbCatalog.Clear;
      statMainInfoMessage.Caption := '   Building catalog...';
      for i := 0 to CurrentDisk.Info.FileCount - 1 do begin
         // Construct Starting Segment List
         j := 0;  SS := '';
         repeat
            SS := SS
             + inttostr(CurrentDisk.Directory[i].SegmentList[j].StartSeg)
             + ','
             + inttostr(CurrentDisk.Directory[i].SegmentList[j].EndOffset);
            inc(j);
            SS := SS + #32;
         until ord(CurrentDisk.Directory[i].SegmentList[j].EndOffset) = 0;
         //
         lbCatalog.AddLine
           ([CurrentDisk.Directory[i].FileName,
             inttostr(CurrentDisk.Directory[i].SectorCount),
             DecodeType(CurrentDisk.Directory[i].StatusFlag,
                        CurrentDisk.Directory[i].Recordlength),
             SS,
             inttostr(CurrentDisk.Directory[i].Address)]);
         end;
      //
      statMainInfoMessage.caption :=
         inttostr(CurrentDisk.Info.SectorCount) + ' sectors - '
         + inttostr(CurrentDisk.Info.FileCount) + ' files';
      statMainInfoMessage2.Caption := OpenImage.FileName;
      statSecondaryInfoMessage1.Caption := 'TI Diskette Image in Memory';
      //
      PagesTI.ActivePage := tabImageInfo;
      //
      DisplayTISector(0, CurrentDisk.SectorList[0].Contents);
      end;
    //
    end;
  screen.Cursor := crDefault;
  lbCatalog.SetFocus;
  lbCatalog.ItemIndex := 0;
  //gridDiskBitMap.Repaint;             //<<-- not strictly necessary?
end;

procedure TfrmMain.actExtractFileExecute(Sender: TObject);
//**********************************
//* Write a file which has been    *
//* extracted from a disk image    *
//* as a stand-alone file.         *
//**********************************
var i, j: integer;
    TempName: string;
    ToF: file;
    NumWritten: Integer;
begin
  tempname := CurrentFile.Name;
  for i := 0 to length(TempName) - 1 do
    case ord(TempName[i]) of
      58: TempName[i] := #95;
      3..47: TempName[i] := #95;
      // etc.
      else TempName[i] := TempName[i];
      end; // case
   //
   SaveDialog.Filter := 'All Files|*.*';
   SaveDialog.DefaultExt := '';
   SaveDialog.FileName := TempName;
   SaveDialog.Title := 'Save formatted text as RTF';
   if SaveDialog.Execute then begin
     AssignFile(ToF, SaveDialog.FileName);
       try
       ReWrite(ToF, 1);
       BlockWrite(ToF, TIFullHeaderBuf, Sizeof(TIFullHeaderBuf), NumWritten);
       for j := 0 to CurrentFile.Header.Sectors do
         BlockWrite(ToF, CurrentFile.SectorList[j].Contents,
                    SizeOf(CurrentFile.SectorList[j].Contents), NumWritten);
       finally
       CloseFile(Tof);
       end; // try
     end; //if
  lblTIFileName.caption := CurrentFile.Name + ' SAVED';
  statMainInfoMessage.Caption := ExtractFileName(SaveDialog.FileName)
    + ' stored in ' + inttostr(CurrentFile.Header.Sectors) + ' sectors';
  pagesTI.ActivePage := tabFileInfo;
end;

procedure TfrmMain.actDisplayfileExecute(Sender: TObject);
//**********************************
//* Extract and display a file     *
//* from a DOAD disk image         *
//**********************************
var Cluster,
    FileSeg,
    DiskSeg: integer;
    Ss: string;
    i, j: integer;
begin
  if lbCatalog.ItemIndex = -1 then begin
    ShowMessage('Please Select a File.');
    exit;
    end;
  statMainInfoMessage.Caption := '';
  //
  // de-allocate sector array
  //
  ClearCurrentFile;
  ClearTIFILESFileInfo;
  //ClearDisketteInfo;     //<-- do not
  //ClearCurrentDiskette   //<-- do not
  ClearMemos;
  //
  CurrentFile.Name := lbCatalog.ItemPart(lbCatalog.ItemIndex, 0);
  CurrentFile.StandAlone := False;
  CurrentFile.SectorsRead := False;
  MakeTIFILESHeader(CurrentFile);
  //
  //  Display header information
  //
  lblTIFileType.caption := CurrentFile.Header.FileType;
  //
  rgProtected.Enabled := true;
  rgProgram.Enabled := true;
  //
  if CurrentFile.Header.ProtectBit then rgProtected.ItemIndex := 0
    else rgProtected.ItemIndex := 1;
  //
  if not CurrentFile.Header.ProgType then begin
    rgDataType.Visible := True;
    rgDataSize.Visible := True;
    label1.Visible := True;
    label3.Visible := True;
    edRecSizeHex.Visible := True;
    edRecSize.Visible := True;
    if CurrentFile.Header.DataSize = 'VAR' then begin
      label2.Visible := True;
      edRecCountHex.Visible := True;
      edRecCount.Visible := True;
      end;
    edRecPerSectorHex.Visible := True;
    edRecPerSector.Visible := True;
    rgProgram.ItemIndex := 1;
    if CurrentFile.Header.datasize = 'FIX' then
      rgDataSize.itemindex := 0 else rgDataSize.itemindex := 1;
    if CurrentFile.Header.dataType = 'DIS' then
      rgDataType.itemindex := 0 else rgDataType.itemindex := 1;
    edRecSizeHex.Text := inttohex(CurrentFile.Header.RecSize, 4);
    edRecCountHex.Text := inttohex(CurrentFile.Header.RecCount, 4);
    edRecPerSectorHex.Text := inttoHex(CurrentFile.Header.RecordsPerSec, 4);
    edRecSize.Value := CurrentFile.Header.RecSize;
    edRecCount.Value := CurrentFile.Header.RecCount;
    edRecPerSector.Value := CurrentFile.Header.RecordsPerSec;
    end
  else rgProgram.ItemIndex := 0;
  //
  edTISectorCountHex.Text := inttoHex(CurrentFile.Header.Sectors, 4);
  edEOFOffsetHex.Text := inttoHex(CurrentFile.Header.EOFOffset, 4);
  edTISectorCount.Value := CurrentFile.Header.Sectors;
  edEOFOffset.Value := CurrentFile.Header.EOFOffset;
  //
  //  Re-allocate Sector Array and
  //    Reset rest of CurrentFile Record
  //
  SetLength(CurrentFile.SectorList, CurrentFile.Header.Sectors);
  CurrentFile.SectorsRead := false;
  CurrentFile.PCName := '';
  lblTIFileName.caption := CurrentFile.Name + ' NOT SAVED';
  ckTIFILESHeader.Checked := True;
  //
  //     now read sectors by iterating through
  //          well not exactly iterating...
  //          maybe we go backwards?
  //
  FileSeg := 0;
  for Cluster := 0 to 74 do begin
    // if at first empty cluster record then break
    if CurrentDisk.Directory[lbCatalog.ItemIndex]
           .SegmentList[Cluster].StartSeg
            = 0 then break;
    //
    DiskSeg :=
       CurrentDisk.Directory[lbCatalog.ItemIndex]
        .SegmentList[Cluster].StartSeg;
    repeat
      statMainInfoMessage.Caption :=
         'Reading Sector ' + inttostr(DiskSeg)
           + ' into File Sector ' + inttostr(FileSeg);
      Move(CurrentDisk.SectorList[DiskSeg].Contents,
        CurrentFile.SectorList[Fileseg].Contents[0],
        sizeof(CurrentDisk.SectorList[DiskSeg].Contents));
      //
      inc(DiskSeg);
      inc(FileSeg);
    until
      FileSeg =
       CurrentDisk.Directory[lbCatalog.ItemIndex]
         .SegmentList[Cluster].EndOffset + 1;
    //
    end; //for which increments Cluster
  //
  //
  //  Need Sectors in memory for these checks
  //
  if CurrentFile.Header.ProgType then
    CurrentFile.BasicFlag := CheckforBasic;
  if not CurrentFile.Header.ProgType then begin
    CurrentFile.CompanionFlag := CheckforComp;
    CurrentFile.TIWriterFlag := CheckforWriter;
    end;
  //
  //  Check boxed before setting SectorsRead
  //
  ckBASIC.Checked := CurrentFile.BASICFlag;
  ckTIWriter.Checked := CurrentFile.TIWriterFlag;
  ckCompanion.Checked := CurrentFile.CompanionFlag;
  //
  //  Sectors now in memory
  //
  CurrentFile.SectorsRead := true;
  statMainInfoMessage.Caption := 'Read ' + inttostr(FileSeg + 1)
           + ' Sectors';
  if CurrentFile.SectorsRead then begin
    ckBASIC.Enabled := CurrentFile.SectorsRead and
        (Copy(CurrentFile.Header.FileType, 1, 5) = 'IMAGE');
    ckCompanion.Enabled := CurrentFile.SectorsRead and
        (Copy(CurrentFile.Header.FileType, 1, 5) = 'IV254');
    ckTIWriter.Enabled := CurrentFile.SectorsRead and
        (Copy(CurrentFile.Header.FileType, 1, 5) = 'DV 80');
  //
  //   create 128 byte string header
  //
  TIFullHeaderBuf[1] := chr(7);
  SS := 'TIFILES';                           // prove me wrong
  Move(SS[1], TIFullHeaderBuf[2], Length(SS));
  //
  TIFullHeaderBuf[9] := chr(CurrentFile.Header.Sectors div 256);
  TIFullHeaderBuf[10] := chr(CurrentFile.Header.Sectors mod 256);
  TIFullHeaderBuf[11] := chr(CurrentFile.Header.StatusFlag);
  TIFullHeaderBuf[12] := chr(CurrentFile.Header.RecordsPerSec);
  TIFullHeaderBuf[13] := chr(CurrentFile.Header.EOFOffset);
  TIFullHeaderBuf[14] := chr(CurrentFile.Header.RecSize);
  TIFullHeaderBuf[16] := chr(CurrentFile.Header.RecCount div 256);
  TIFullHeaderBuf[15] := chr(CurrentFile.Header.RecCount mod 256);
  for i := 17 to 128 do TIFullHeaderBuf[i] := chr(0);
  //
  //  Display 16 bytes of Header
  //
    for i := 0 to 1 do
      for j := 0 to 7 do
       frmMain.gridHeader.Cells[j, i] :=
         format(   '%.2x', [  ord(TIFullHeaderBuf[(i * 8) + j + 1])  ]   );
  //
  //   Then Display first Sector
  //
  DisplayingDiskImage := False;
  DisplayingDiskFile := False;
  DisplayingFile := True;
  DisplayTISector(0, CurrentFile.SectorList[0].Contents);
  PagesTI.ActivePage := tabFileInfo;
  end;
end;


//***********************************
//* Display and Conversion Actions  *
//***********************************


procedure TfrmMain.actFormatCompanionExecute(Sender: TObject);
//*****************************************************
//*                                                   *
//*  The 'companion' word processor used a custom     *
//*  ondisk format with character substitutions for   *
//*  spaces, linefeeds, etc. The file format is IV254 *
//*  so the first sector usuallt begins >FE for the   *
//*  record length, but the next byte is the          *
//*  companion record length, always one less. From   *
//*  there each record is a stream of text,           *
//*  substituted characters and control characters.   *
//*  End-of-Sector is marked by HEX FF.               *
//*                                                   *
//*  I translate it here using a method loosely based *
//*  on my old RITEFRND:A TI Basic program, which     *
//*  translated between companion and TI-Writer       *
//*  and vice-versa.                                  *
//*                                                   *
//*  Formatting for dot-matrix printers was pretty    *
//*  crude anyway, so we only handle and simulate     *
//*  the obvious stuff here. Besides, I long since    *
//*  lost my old 'companion' docs, and have created   *
//*  most of the stuff here from trial-and-error      *
//*  examiniation of sample texts.                    *
//*                                                   *
//*****************************************************
var SecNum, SecIndex, RecLen: integer;
    var ParIndent, OldParIndent, TrapCount, TrapType, ParamCode: integer;
    TempString, TranString, CodeStr, ParamChar, ParamString: string;
    SelAttrLineFlag, CenterLine, BoldFlag, UnderFlag, DoNotEat, NoText: boolean;
    LMargin, RMargin, FormWidth, LineSpacing, MidLine: integer;
  //
  //  internal utility procedure to advance file pointers
  //
  procedure NextByte(Count: integer);
  begin
   if SecIndex < 256 - Count then
     SecIndex := SecIndex + Count
   else begin                //start of new sector
     SecNum := SecNum + 1;
     SecIndex := 0;
     end;
  end;
  //
  //  internal utility procedure Next Line
  //
  procedure NextLine;
  begin
    if not SelAttrLineFlag then begin
      memFormatted.Lines.add(tempstring);
      end
    else begin
      SelAttrLineFlag := False;
      // add the balance of the line
      memFormatted.SelText := TempString;
      memFormatted.Lines.Add('');
      end;
  end;
  //
  //  internal utility procedure apply paragraph setting
  //
  procedure ApplySettings;
  begin
    //with memFormatted.Paragraph do begin
    with memFormatted do begin
      if CenterLine then Alignment := taCenter else Alignment := taLeftJustify;
      //FirstIndent := 4 * (LMargin + Parindent);    //FIXME
      //LeftIndent :=  - (4 * ParIndent);
      //RightIndent :=  4 * RMargin;
      //SetLineSpacing(lsrMultiple, LineSpacing);
      end; //with
  end;
  //
begin
  //
  //    conversions used in RITEFRND:A
  //    ------------------------------
  //    companion chr           TIW char
  //    128                     32   SPACE
  //    129                     13   CR   // Ctl-C - Center Line
  //    130     -------->       10   LF
  //    130     <--------       13   CR   // Ctl-L - LineFeed
  //    131                     13   CR   // Ctl-P - Paragraph
  //    132                     13   CR   // Ctl-M - MidLine (?para indent?)
  //    136                     12   FF   // Ctl-N - NewPage
  //    137                     94   ^    // Ctl-J - (a 'required' space)
  //
  //    additional codes not in RITEFRND:A
  //    ----------------------------------
  //
  //    026                     Slash Left    ?
  //    027                     Slash Right   ?

  //    133                     CTL-9    start of underline (we use italics)
  //    134                     CTL-0    end of underline
  //
  //    135                     CTL-I  - Might be ......?
  //
  //    138                     CTL-T  - no ascii param but no idea of meaning
  //
  //    139                     CTL-H  - Horizontal Tab
  //
  //    140                     REVISE - which can be followed
  //                                     by a number of characters
  //                                     and needs to be interpreted
  //                                     first alpha after must be
  //                                     in proper subset. first
  //                                     non-numeric after the alpha
  //                                     is straight text.
  //
  //    141                     CTL-A  - which can be followed
  //                                     by literal numeric characters
  //                                     representing an ASCII code
  //                                     sent straight to the printer
  //
  //    142                     CTL-G  - Graphics Character (ascii)
  //
  //    143                     CTL-B  - Might be overstrike in Epson?
  //
  //  **************************************************************
  //
  //  we're just going to build paragraphs, not lines,
  //  so we let the control do the wrapping
  //
  memFormatted.WantTabs := true;
  memFormatted.WordWrap := true;
  memFormatted.ScrollBars := ssVertical;
  memFormatted.Clear;
  //
  //  some companion 'defaults'
  //
  ParIndent := 5;
  LMargin := 8;
  RMargin := 8;
  FormWidth := 80;
  MidLine := 32;
  LineSpacing := 1;
  CenterLine := False;
  // apply them
  ApplySettings;
  //
  //  Lower the Trap Flags
  //
  SelAttrLineFlag := False;
  BoldFlag := False;          // not used
  UnderFlag := False;         // not used
  DoNotEat := False;
  //
  NoText := True;
  //
  //  Clear the traps
  //
  TrapType := 0;
  TrapCount := 0;
  TempString := '';
  TranString := '';
  CodeStr := '';
  ParamChar := '';
  ParamString := '';
  //
  Screen.Cursor := crHourGlass;
  //
  //  start of byte-by-byte conversion
  //
  for SecNum := 0 to CurrentFile.Header.Sectors - 1 do begin
    SecIndex := 1;
    statMainInfoMessage.Caption := '   converting sector '
       + inttostr(SecNum) + ' byte ' + inttostr(SecIndex);
    //
    // Read IV254 true Length Byte - never really used
    //
    RecLen := ord(CurrentFile.SectorList[SecNum].Contents[SecIndex]);
    //
    // skip companion length byte which is always one less
    //
    NextByte(1);
    //
    //  Record #0 always starts with 4 bytes of character counting
    //
    //if SecNum = 0 then NextByte(4);
    //
    //  Other long files (parts of multi-part or chained documents) have an
    //       even longer header, but I'm going to blissfully ignore that
    //
    //
    while CurrentFile.SectorList[SecNum].Contents[SecIndex] <> #255 do begin
      //
      //  Special Cases
      //
      if TrapCount > 0 then begin

        //
        //  in-trap processing
        //
        Case TrapType of
          // ASCII Escape Code
           0: begin
              if ((ord(CurrentFile.SectorList[SecNum].Contents[SecIndex])
                     >= 48)
                 and (ord(CurrentFile.SectorList[SecNum].Contents[SecIndex])
                     <= 57))
                  then inc(TrapCount)
                  else DoNotEat := true;
              end;
          // Graphics ASCII code
          1: begin
             if ((ord(CurrentFile.SectorList[SecNum].Contents[SecIndex])
                     >= 48)
                 and (ord(CurrentFile.SectorList[SecNum].Contents[SecIndex])
                     <= 57)) then begin
               Inc(TrapCount);
               CodeStr := CodeStr +
                 CurrentFile.SectorList[SecNum].Contents[SecIndex];
               end
               else begin
                  TempString := Tempstring + chr(strtoint(CodeStr));
                  DoNotEat := true;
                  end;
             end;
          // Revise    65-90 = alpha
          2: begin
             if TrapCount = 99 then begin
                ParamChar := CurrentFile.SectorList[SecNum].Contents[SecIndex];
                TrapCount := 2;
                end
             else begin  //we already have param so we're looking for numerics
                if ((ord(CurrentFile.SectorList[SecNum].Contents[SecIndex])
                         >= 48)
                  and (ord(CurrentFile.SectorList[SecNum].Contents[SecIndex])
                         <= 57)
                  and (ord(CurrentFile.SectorList[SecNum].Contents[SecIndex])
                         <> 45))
                  then begin
                  ParamString := ParamString
                      + CurrentFile.SectorList[SecNum].Contents[SecIndex];
                  inc(TrapCount);
                  end
                  else begin
                  if paramstring <> '' then
                     ParamCode := strtoint(Paramstring)
                  else ParamCode := 0;
                  //
                  //  Act on (certain) Parameters  (no error checking!!!)
                  //
                   case ord(ParamChar[1]) of
                     //LMargin
                     76, 108: if ParamCode = 0 then
                                    LMargin := 8
                              else LMargin := ParamCode;
                     //Midline
                     77, 109: if ParamCode = 0 then
                                    Midline := 32
                              else MidLine := ParamCode;
                     //ParIndent
                     80, 112: if ParamCode = 0 then
                                    ParIndent := 5
                              else ParIndent := ParamCode;
                     //RMargin
                     82, 114: if ParamCode = 0 then
                                    RMargin := 8
                              else RMargin := ParamCode;
                     //Line Spacing 
                     83, 115: if ParamCode = 0 then
                                    LineSpacing := 1
                              else LineSpacing := ParamCode;
                   end;
                  //
                  DoNotEat := true;
                  end;
                end;
             end;
          // UnderLine
          3: begin
               if (ord(CurrentFile.SectorList[SecNum].Contents[SecIndex])
                <> 134) then begin
                  //
                  //  inside trap we handle spaces only!!!!!
                  //
                  case ord(CurrentFile.SectorList[SecNum].Contents[SecIndex]) of
                    128: TempString := TempString + #32;
                    137: TempString := TempString + #32;
                  else tempstring := tempstring
                    + CurrentFile.SectorList[SecNum].Contents[SecIndex];
                  end;
                  inc(TrapCount);
                  end
                else begin
                  memFormatted.SelText := TempString;
                  TempString := '';
                  memFormatted.SelAttributes.Style := [];
                  SelAttrLineFlag := True;
                  end;
              end;
          end;  //case
        //
        // always
        //
        Dec(TrapCount);
        if not DoNotEat then NextByte(1) else DoNotEat := False;
        end
      //
      //  normal processing
      //
      else begin
        case ord(CurrentFile.SectorList[SecNum].Contents[SecIndex])of

           0: ;  // no op

           //
           //  the bold slashes ??
           //

           26: ;

           27: ;

           //
           //  'normal' text, not within one of the traps below
           //
           32 .. 127: begin
                       if NoText then NoText := False;
                       tempstring := tempstring    // straight ASCII
                       + CurrentFile.SectorList[SecNum].Contents[SecIndex];
                       end;
           //
           //  various conversions, and special traps to read the
           //     parameters of certain control codes and apply
           //     to memFormatted.SelAttributes
           //

           128: tempstring := tempstring + #32;        // the space character

           //
           //   We're going to center as formatted Rich Text, rather than
           //   counting the form width, line width, subtracting, etc...
           //
           129: begin                                  // Center Line
                  //
                  //  only if not at top of file????
                  //
                  if not NoText then NextLine;
                  //
                  CenterLine := True;                  // we set this flag
                  ApplySettings;                       // Set up for Next Para
                  CenterLine := False;                 // Clear the Flag
                  tempstring := '';                    // Clear the TempString
                  end;

           130: begin                                  // LineFeed
                  NextLine;
                  //
                  OldParIndent := ParIndent;           // Save Value
                  ParIndent := 0;                      // temp change
                  ApplySettings;                       // Set up for Next Para
                  ParIndent := OldParIndent;           // restore Value
                  tempstring := '';                    // Clear the TempString
                  end;

           131:  begin                                 // Paragraph
                  NextLine;
                  //
                  ApplySettings;                       // Set up for Next Para
                  memFormatted.Lines.add('');          // we add extra line
                  tempstring := '';                    // Clear the TempString
                  end;

           132:  begin                                 // MidLine Indent
                  NextLine;
                  //
                  ApplySettings;                       // Set up for Next Para
                  tempstring
                    := stringofchar(#32, MidLine);     // we honor MidLine
                  end;

           //
           //  We Italicize rather than underline.
           //

           133: begin                                   // Underline
                if TempString <> '' then begin
                  memFormatted.SelText := Tempstring;   // dump line so far
                  TempString := '';
                  end;
                memFormatted.SelAttributes.Style := [fsItalic]; //  We Italicize
                SelAttrLineFlag := True;
                TrapType := 3;                          // Set type to underline
                TrapCount := 1;                         // set up trap
                end;

           // since it's caught in the trap we should never see this
           134: ;                                       // end underline

           135: ;                                       // unknown  ?

           136:  begin                                  // New Page
                  NextLine;
                  //
                  ApplySettings;                        // Set up for Next Para
                  tempstring := '';
                  end;

           // Ctl-J
           137: tempstring := tempstring + #32;         // a 'required' space
           // Ctl-T
           138: ;                                       // unknown
           // Ctl-H
           139: tempstring := tempstring + #9;          // Horizintal tab

           //
           //   The revise command... we have to parse the code and paramater(s)
           //      and take appropriate action for each. or we store general
           //      formatting and apply it only when we add the tempstring
           //
           //      Code   Min     Max     Default
           //      W      40      132     80      form width
           //      F      22      132     66      form length
           //      L      0       35      8       left margin
           //      R      0       25      8       right margin
           //      T      2       34      6       top margin
           //      B      2       20      6       bottom margin
           //      S      1       5       1       (line) spacing
           //      P      -20     20      5       Paragraph Indent
           //      M      -20     60      32      Midline Indentation
           //                                      (not true centering!)
           //      H      1       8       1       Horizontal Tab Group
           //                                      (tabs not defined in the
           //                                       document, a weakness)
           //      J      0       9       0       Justification
           //                                      (never implemented)
           //
           //
           140: begin;
                 TrapType := 2;
                 ParamChar := '';                  // Parameter Code
                 ParamString := '';                // Parameter Value
                 TrapCount := 99;                  // artificial count as flag
                 end;


           141: begin                              // ASCII (printer escape)
                 TrapType := 0;
                 TrapCount := 1;                   // look for the ascii
                 end;


           142: begin;                             // GRAPHICS ascii code
                 TrapType := 1;
                 TrapCount := 1;
                 CodeStr := '';
                 end;

           143: ;                                  // unknown Ctl-B

           255: Break;                             // end of sector marker

           else
           //
           // everything else
           // is really unknown so
           // just skip the byte
           //
             tempstring := tempstring;
           end;  // case
        //
        NextByte(1);
        end; //else
      end;
    //
    end; //of document
  //
  NextLine;
  //
  if CenterLine then CenterLine := False;
  TempString := '';
  //
  //  we got here so
  //
  CurrentFile.FormattedFlag := true;
  //

  if CurrentFile.FormattedFlag then begin

    statMainInfoMessage.Caption := '  conversion complete';
    pagesTI.ActivePage := tabFormatted;
    memFormatted.SetFocus;
    SendMessage(MemFormatted.Handle,EM_SETSEL, 0, 0);
    end
  else
    statMainInfoMessage.Caption := '  conversion cancelled';
  screen.Cursor := crDefault;
end;

procedure TfrmMain.actDisplayTextExecute(Sender: TObject);
//*****************************************************
//*  DV80, DVany, DFix Display, formatted or not      *
//*                                                   *
//*  TI Writer is essentially a line-by-line text     *
//*  editor  which pre-formats its own output,        *
//*  so I do not bother with constructing             *
//*  formatted paragraphs. I'll try to note and       *
//*  capture italic and bold (underscore and over-    *
//*  strike) codes, and reflect the most basic line   *
//*  formatting, like spacing and indentation.        *
//*****************************************************
var SecNum, SecIndex, RecLen, i: integer;
    TIUnderscoreStart, TIOverstrikeStart: integer;
    TIUnderscoreStop, TIOverstrikeStop: integer;
    FormWidth, CenterCount, LMargin, RMargin, ParIndent, LineSpacing: integer;
    TempString: string;
    TIUnderScoreFlag, TIOverStrikeFlag, TICenterLineFlag: boolean;
  //***************************************************
  // some subfunctions for TI Writer formatting       *
  //     only get called when TIWriter box is checked *
  //***************************************************
  //*  TI Writer sub procedure to format and add      *
  //*  each individual line                           *
  //***************************************************
  procedure AddLine(var s: string);
  var i, j: integer;
  t: string;
  begin
  TIUnderScoreStart := -1;
  TIUnderScoreStop := -1;
  TIOverStrikeStart := -1;
  TIOverStrikeStop := -1;
  //
  //TIUnderScoreFlag := False;
  //TIOverStrikeFlag := False;
  t := '';
  //
  //  parse the line
  //
  for i := 1 to length(s) do begin
    case ord(s[i]) of
      {13: begin
           if TIOverstrikeflag then begin
            TIOverstrikeFlag := False;
            TIOverstrikeStop := i - 1;
            end
           else if TIUnderScoreFlag then begin
            TIUnderscoreFlag := False;
            TIUnderScoreStop := i - 1;
            end;
           end; }
      30: ;                                // no op
      31: ;                                // no op
      32: begin
           if TIOverstrikeflag then begin
            TIOverstrikeFlag := False;
            TIOverstrikeStop := i;
            end
           else if TIUnderScoreFlag then begin
            TIUnderscoreFlag := False;
            TIUnderScoreStop := i;
            end;
           t := t + s[i];
           end;
      33..37: t := t + s[i];
      38: begin
           TIUnderScoreFlag := True;
           TIUnderscoreStart := i;
           end;
      39..63: t := t + s[i];
      64: begin
           TIOverStrikeFlag := True;
           TIOverStrikeStart := i;
           end;
      65..93: t := t + s[i];
      94: if TIOverStrikeFlag or TIUnderScoreFlag
             then t:= t + #32
           else t := t+ s[i];
      95..126: t := t + s[i];
      end; //case
    end; // for
   TIOverstrikeFlag := False;
   TIOverstrikeStop := length(t);
   TIUnderscoreFlag := False;
   TIUnderScoreStop := length(t);
   //
   for j := 1 to LineSpacing do memformatted.lines.Add('');
   if CenterCount > 0 then begin
     Memformatted.SelText := StringofChar(#32,
            (FormWidth - length(t)) div 2);
     Dec(CenterCount);
     end;
   //
   for i := 1 to length(t) do begin
       //
       //  check for formatting turn-on
       //
       if i = TIOverStrikeStart then
          memformatted.SelAttributes.Style :=
            memformatted.SelAttributes.Style + [fsBold];
       if i = TIUnderscoreStart then
          memformatted.SelAttributes.Style :=
            memformatted.SelAttributes.Style + [fsItalic];
       //
       //  add char
       //
       memFormatted.Seltext := memFormatted.SelText + t[i];
       //memFormatted.Seltext := t[i];  //old form with RichEdit98
       //
       //  check for turn-off
       //
       if i = TIOverStrikeStop then
          memformatted.SelAttributes.Style :=
            memformatted.SelAttributes.Style - [fsBold];
       if i = TIUnderscoreStop then
          memformatted.SelAttributes.Style :=
            memformatted.SelAttributes.Style - [fsItalic];
       //
       end;
   end; //sub procedure
   //* ***************************************************
   //*  sub function to process line for TIWriter codes  *
   //*****************************************************
   function ProcessTIWriterLine(var s: string): boolean;
   begin
     Result := True;
     if copy(s, 1, 1) = '.' then begin
       Result := False;
       //  ShowMessage(s);
       //  ShowMessage(Copy(s, 4, length(s) - 3));
       //
       //  act on formatting
       //
       if copy(s, 1, 3) = '.CE' then
         if length(s) > 3 then
           try
           CenterCount := strtoint(trim(Copy(s, 4, length(s) - 3)))
           except
           CenterCOunt := 1;
           end;
       if copy(s, 1, 3) = '.LM' then
         if length(s) > 3 then
           try
           LMargin := strtoint(trim(Copy(s, 4, length(s) - 3)))
           except
           LMargin := 8;
           end;
       if copy(s, 1, 3) = '.RM' then
         if length(s) > 3 then begin
           try
           RMargin := strtoint(trim(Copy(s, 4, length(s) - 3)))
           except
           RMargin := 8;
           end;
           //
           //  this is a kludgey approximation
           //
           FormWidth := RMargin - LMargin;
           end;
       if copy(s, 1, 3) = '.LS' then
         if length(s) > 3 then
           try
           LineSpacing := strtoint(trim(Copy(s, 4, length(s) - 3)))
           except
           LineSpacing := 1;
           end;
       if copy(s, 1, 3) = '.IN' then
         if length(s) > 3 then
           try
           ParIndent := strtoint(trim(Copy(s, 4, length(s) - 3)))
           except
           ParIndent := 5;
           end;
       end; //if
   //ShowMessage('LeftMargin = ' + inttostr(LMargin) + #10 + #13 +
   //            'RightMargin = ' + inttostr(RMargin) + #10 + #13 +
   //            'ParIndent = ' + inttostr(ParIndent) + #10 + #13 +
   //            'LineSpacing = ' + inttostr(LineSpacing) + #10 + #13 +
   //            'CenterCount = ' + inttostr(CenterCount));
   end; //sub function

//
//  main procedure code
//

begin
//*****************************************************
//*                                                   *
//*  Translating TI_Writer and other DV80 files       *
//*  into straight PC text is relatively simple.      *
//*  VAR records have use length byte for each record *
//*  followed by the actual data. VAR files also use  *
//*  a length byte of >FF to mark end of data in a    *
//*  given sector. One just reads the lines and       *
//*  stacks them up.                                  *
//*                                                   *
//*  If the TIWriter File flag is set we look for     *
//*  and act on the 'dot' codes and parse the lines   *
//*  for underscores, circumflex marks and other      *
//*  TIWriter specific stuff. Formatting for dot-     *
//*  matrix printers was pretty crude anyway, so we   *
//*  only handle and simulate the obvious stuff here. *
//*                                                   *
//*  For DF records, we just do the math. There       *
//*  is no end of sector marker, so if another record *
//*  won't fit we just skip to the next sector.       *
//*  Otherwise it's s similar process to DV80 files.  *
//*  Unfortunately, we usually get some trash         *
//*  text at the end of the FIX file when there is    *
//*  a lot of unused space in the last sector.        *
//*                                                   *
//*****************************************************
  //
  TIOverStrikeFlag := False;
  TIUnderScoreFlag := False;
  TICenterLineFlag := False;    // not used
  //
  ParIndent := 5;
  LMargin := 8;
  RMargin := 8;
  CenterCount := 0;
  FormWidth := 80;
  LineSpacing := 1;
  //
  //  we're doing line-by line formatting here, so no wrap
  //
  memFormatted.WantTabs := true;
  memFormatted.WordWrap := False;
  memFormatted.ScrollBars := ssBoth;
  memFormatted.Clear;
  //
  Screen.Cursor := crHourGlass;
  //  DV80 TIWriter type files
  if CurrentFile.Header.DataSize = 'VAR' then begin
    //  VARiable routine
    for SecNum := 0 to CurrentFile.Header.Sectors - 1 do begin
      statMainInfoMessage.Caption := '   converting sector '
                 + inttostr(SecNum) + '..........';
      // reset index
      SecIndex := 0;
      // read 1st length byte
      RecLen := ord(CurrentFile.SectorList[SecNum].Contents[SecIndex]);
      //  read lines
      while ((RecLen <> $FF) and (recLen <> 0)) do begin
        // $FF is end of sector marker
        // increment sector pointer
        inc(SecIndex);
        // read line (record)
        TempString := '';
        //
        for i := 1 to RecLen do begin
          TempString := TempString
            + CurrentFile.SectorList[SecNum].Contents[SecIndex];
          inc(SecIndex);
          end;
        //
        //  we have the line now what do we do with it?
        //
        if (not CurrentFile.TIWriterFlag) then
          //
          //  We're not formatting so just dump the line
          //
          memFormatted.Lines.Append(TempString)
          //
          //  if we are formatting then test the line
          //
        else if ProcessTIWriterLine(tempstring) then
          //
          //  if it's a 'real' text line and
          //     not a formatting dot-command then
          //
          AddLine(TempString);
        //
        // read next length byte to get ready for next line
        //
        RecLen := ord(CurrentFile.SectorList[SecNum].Contents[SecIndex]);
        end;  // [while] end of records in Sector
      RecLen := 0;
      end;  // [for] of sectors in file
    CurrentFile.FormattedFlag := true;
    end  // if VARiable
  //
  else begin  //DF files
    if DISFIXWarn.Execute then begin       //Warn about display of object code
      RecLen := CurrentFile.Header.RecSize;
      for SecNum := 0 to CurrentFile.Header.Sectors - 1 do begin
        statMainInfoMessage.Caption := '   converting sector '
                 + inttostr(SecNum) + '..........';
        // reset index
        SecIndex := 0;
        while (256 - SecIndex) > RecLen do begin
          TempString := '';
          // read line (record)
          for i := 1 to RecLen do begin
            TempString := TempString
              + CurrentFile.SectorList[SecNum].Contents[SecIndex];
            inc(SecIndex);
            end;
          // Append Output
          try memFormatted.Lines.Append(TempString); except end;
          end;  // [while] end of records in Sector
        end;  // [for] of sectors in file
      CurrentFile.FormattedFlag := true;
      end; //if warn.execute
    end; // else
  //
  if CurrentFile.FormattedFlag then begin
    statMainInfoMessage.Caption := '  conversion complete';
    pagesTI.ActivePage := tabFormatted;
    memFormatted.SetFocus;
    SendMessage(MemFormatted.Handle,EM_SETSEL, 0, 0);
    end
  else
    statMainInfoMessage.Caption := '  conversion cancelled';
  screen.Cursor := crDefault;
end;

procedure TfrmMain.actParseBASICExecute(Sender: TObject);
  //***********************************************************
  //*                                                         *
  //*   The ondisk format of BASIC programs mirrors their     *
  //*   in memory storage. That's unfortunate, to say the     *
  //*   least.                                                *
  //*                                                         *
  //*   First come line numbers (sorted decending, which is   *
  //*   perverse) paired with pointers to the matching        *
  //*   console memory addresses where the matching state-    *
  //*   ments were at when the program was SAVed and where    *
  //*   they will be when loaded again.                       *
  //*                                                         *
  //*   Next come the actual lines, tokenized (of course)     *
  //*   with a length byte preceeding the line, and a >00     *
  //*   to mark the line ending, as well. The statements      *
  //*   are in MEMORY ORDER, not sequential. Isn't that       *
  //*   interesting? so you hava to put the statements in     *
  //*   the same RELATIVE ORDER as the memory addresses of    *
  //*   the line numbers, and then of course, reverse the     *
  //*   overall order ot the line-number-statement com-       *
  //*   binations. A system only a computer could love!       *
  //*                                                         *
  //*   De-tokenizing is also not exactly straightforward:    *
  //*                                                         *
  //*   There's a special token to indicates that the next    *
  //*   XX bytes are a literal string, and that token is      *
  //*   followed by a length byte for the literal string.     *
  //*   The literal strings are all the non-tokenized code.   *
  //*                                                         *
  //*   There's another token to indicate a quoted string.    *
  //*   That token must be translated into a double-quote     *
  //*   character, then comes the length byte, and at the     *
  //*   end of the quoted string the closing double-quote     *
  //*   is only implied, so to list the program correctly     *
  //*   the closing double-quote must be suppplied.           *
  //*                                                         *
  //*   Another special token indicates that what follows     *
  //*   is a line number address, so one has to eat that      *
  //*   token, read the next two byte word and treat it as    *
  //*   a line number.                                        *
  //*                                                         *
  //*   Very messy.                                           *
  //*                                                         *
  //***********************************************************
var BasicLineNumbers, BasicStatements, BasicStatementAddress: TStrings;
//    LineNumBlock: array[0..3] of char;
    ALine, DebugLine: string;
    OldLineNumVal, LineNumVal, QuotedStringLength: integer;
    StatementNum, FileNum, ACell: integer;
    LineCount, SecNum, SecIndex, ThisStatementLength, i, j: integer;
    TempChar, TempAddress, TempLineNumber, TempStatement: string;
  //
  //  Utility Procedure to Advance File Pointers
  //
  procedure NextByte(Count: integer);
  begin
   if SecIndex < 256 - Count then
     SecIndex := SecIndex + Count
   else begin                //start of new sector
     SecNum := SecNum + 1;
     SecIndex := 0;
     end;
  end;
  //
  //  main procedure code
  //
begin
  memFormatted.WantTabs := true;
  memFormatted.WordWrap := False;
  memFormatted.ScrollBars := ssVertical;
  memFormatted.Clear;
  //memFormattedNew.lines.Clear;
  //
  BasicLineNumbers := TStringList.Create;
  BasicStatements := TStringList.Create;
  BasicStatementAddress := TStringList.Create;
  //
  Screen.Cursor := crHourGlass;
  //
  QuotedStringLength := 0;
  OldLineNumVal := 40000; // dummy line number
  SecNum := 0;            // start at the top
  SecIndex := 8;          // reset index to start of Basic Line Number Table
  //
  // Construct Line Number Table
  //
  while true do begin
    // read line Number byte
    LineNumVal :=
     ((256 * ord(CurrentFile.SectorList[SecNum].Contents[SecIndex]))
           + ord(CurrentFile.SectorList[SecNum].Contents[SecIndex + 1]));
    // as long as line numbers decrease
    if LineNumVal < OldLineNumVal then begin
     // then read line number and memory address
     OldLineNumVal := LineNumVal;
     BasicLineNumbers.Append(format('%5d', [LineNumVal]));
     BasicStatementAddress.Append(format('%5d',
      [((256 * ord(CurrentFile.SectorList[SecNum].Contents[SecIndex + 2]))
            + ord(CurrentFile.SectorList[SecNum].Contents[SecIndex + 3]))- ProgramTableStart]));
     NextByte(4);
     end
    else break;  // get out of loop, in the process
                 //  saving Sector Number and Sector Index
  end;
  //      We're now at the start of the statement table.
  //      Each entry is a length byte, followed by the
  //      tokenized statement, followed by a >00 byte.
  //
  LineCount := BasicLineNumbers.Count; // we use this
  for i := 1 to LineCount do begin
    //
    if i <> 1 then NextByte(1); // skip previous line terminating >00
    //
    // Don't really use this value but let's store it anyway
    //
    ThisStatementLength :=
      ord(CurrentFile.SectorList[SecNum].Contents[SecIndex]);
    NextByte(1);                // eat length byte
    //
    //  Detokenize line
    //
    while CurrentFile.SectorList[SecNum].Contents[SecIndex] <> #0 do begin
      ALine := ALine +
        ThisToken(CurrentFile.SectorList[SecNum].Contents[SecIndex]);
      DebugLine := DebugLine +
        CurrentFile.SectorList[SecNum].Contents[SecIndex];
      //
      //  if we're in a quoted string keep track,
      //    check for the end of the string, and
      //      insert close quote if at the end
      //
      if QuotedStringLength > 0 then begin
        dec(QuotedStringLength);
        if QuotedStringLength = 0 then ALine := ALine + '"';
        end;
      //
      //  special behaviors for special tokens
      //
      case ord(CurrentFile.SectorList[SecNum].Contents[SecIndex]) of
         199: begin ;  // save string length bytes
               NextByte(1);
               DebugLine := DebugLine +
                 CurrentFile.SectorList[SecNum].Contents[SecIndex];
               QuotedStringLength :=
                 ord(CurrentFile.SectorList[SecNum].Contents[SecIndex]);
              end;

         200: begin   // non-quoted string so just eat the length byte
               NextByte(1);
               DebugLine :=
                 DebugLine + CurrentFile.SectorList[SecNum].Contents[SecIndex];
              end;

         201: begin   //statement # reference
               NextByte(1);  //eat the >c9
               DebugLine :=
                 DebugLine + CurrentFile.SectorList[SecNum].Contents[SecIndex];
               //
               //    calculate the number
               //
               StatementNum :=
                 256 * ord(CurrentFile.SectorList[SecNum].Contents[SecIndex]);
               NextByte(1);
               DebugLine :=
                 DebugLine + CurrentFile.SectorList[SecNum].Contents[SecIndex];
               StatementNum := StatementNum +
                 ord(CurrentFile.SectorList[SecNum].Contents[SecIndex]);
               //
               //  Add the number to the line
               //
               ALine := ALine + format('%5d', [StatementNum]);
               end;
         end;   // of special cases
      //
      //  Now proceed to next char/token
      //
      NextByte(1);
      end;
    //
    //  add the parsed line to the array of lines
    //
    BasicStatements.Append(trim(ALine));
    ALine := '';
    //
    //  end of parsing loop
    //
    end;
  //
  //   finished capturing line numbers, address offsets and statements
  //
  statMainInfoMessage.Caption :=
  '   Formatted ' + inttostr(Linecount - 1) + ' lines...';
  //
  //  Now fold the arrays and display
  //
  //    I just use a pair of simple bubble sorts, since it's
  //    an easy way to sort multiple arryas by one array.
  //    If you want to make a fansy Shell Sort, be my guest.
  //
  //    first sort the line numbers and address offsets
  //     into order ot the address offsets
  //
  for i := 0 to LineCount - 1 do begin
     for j := i + 1 to LineCount - 1 do begin
        if strtoint(BasicStatementAddress[i])
          > strtoint(BasicStatementAddress[j]) then begin
           //
           tempAddress := BasicStatementAddress[i];
           BasicStatementAddress[i] := BasicStatementAddress[j];
           BasicStatementAddress[j] := tempAddress;
           //
           tempLineNumber := BasicLineNumbers[i];
           BasicLineNumbers[i] := BasicLineNumbers[j];
           BasicLineNumbers[j] := tempLineNumber;
        end; // if
     end; // j
  end; // i
  //
  //    then sort all three arrays
  //     into order ot the line numbers
  //
  for i := 0 to LineCount - 1 do begin
     for j := i + 1 to LineCount - 1 do begin
        if strtoint(BasicLineNumbers[i])
          > strtoint(BasicLineNumbers[j]) then begin
           //
           tempLineNumber := BasicLineNumbers[i];
           BasicLineNumbers[i] := BasicLineNumbers[j];
           BasicLineNumbers[j] := tempLineNumber;
           //
           tempAddress := BasicStatementAddress[i];
           BasicStatementAddress[i] := BasicStatementAddress[j];
           BasicStatementAddress[j] := tempAddress;
           //
           tempStatement := BasicStatements[i];
           BasicStatements[i] := BasicStatements[j];
           BasicStatements[j] := tempStatement;
        end; // if
     end; // j
  end; // i
  //
  //  Now build the output memo
  //
  for i := 0 to LineCount - 1 do begin
    //
    //  I could parse lines to clean double spaces
    //    but that would destroy lines with strings or comments
    //    which include spaces, so the heck with it
    //
    {TempStatement := '';
    j := 1;
    while j <= length(trim(BasicStatements[i])) do begin
      tempchar := copy(trim(BasicStatements[i]), j, 1);
      tempStatement := TempStatement + Tempchar;
      if tempchar = #32 then
        while copy(trim(BasicStatements[i]), j, 1) = #32 do inc(j)
      else
        inc(j);
      end;
    BasicStatements[i] := TempStatement;}
    //
    memFormatted.lines.Add(trim(BasicLineNumbers[i])
     + #32 + Trim(BasicStatements[i]));
    end;
  //
  //  Clean Up
  //
  CurrentFile.FormattedFlag := True;
  //
  Screen.Cursor := crDefault;
  pagesTI.ActivePage := tabFormatted;
  memFormatted.SetFocus;
  SendMessage(MemFormatted.Handle,EM_SETSEL, 0, 0);
  //
  BasicLineNumbers.Free;
  BasicStatements.Free;
  BasicStatementAddress.Free;
end;

//***********************************************
//*  Major Form-Object Published Subroutines    *
//***********************************************

procedure TfrmMain.DisplayTISector(SecNum: integer; Sec: array of char);
var i, gCol, gRow: integer;
begin
  //
  //  byte-by-byte add Buffer to memo and grid and to integer array
  //
  edSectorNumHex.Text := inttohex(SecNum, 2);
  edSectorNum.Value := SecNum;
  MemRawText.Clear;
  gCol := 1;
  gRow := 1;
  ;;
  for i := 0 to 255 do begin
    gridSectors.Cells[gCol, gRow] :=
     format('%.2x',[ord(Sec[i])]);
    gridSectors.Cells[gCol + (gridSectors.ColCount div 2), gRow] := Sec[i];
    //
    memRawText.Text := memRawText.Text + Sec[i];
    //
    inc(gCol);
    if gCol > (gridSectors.ColCount div 2) then begin
      gcol := 1;
      inc(gRow);
      end;
    end;
  CurrentSector := SecNum;
end;

procedure TfrmMain.ReadTIFilesSectors(Sender: TObject);
//**********************************************
//*   Written as separate subroutine           *
//*   for development but unique to stand-     *
//*   alone files and ALWAYS called when       *
//*   one is opened.                           *
//**********************************************
//*   Treats the file as if it were on a       *
//*   256 byte sector disk, and stores it as   *
//*   a set of (continugous) whole sectors     *
//*   which we fetch via blockread, treating   *
//*   the file as a sort of stream.            *
//**********************************************
var i: integer;
  FromF: file;
  NumRead: Integer;
  TIDummyBuf: array[1..128] of Char;
begin
  //
  //  open file and pre-read header 128 bytes
  //
  AssignFile(FromF, CurrentFile.PCName);
  Reset(FromF, 1);
  //
  //  skip header if present
  //
  if CurrentFile.Header.HeaderString = 'TIFILES' then
    BlockRead(FromF, TIDummyBuf, SizeOf(TIDummyBuf), NumRead)
  else begin
    statMainInfoMessage.Caption := 'Not a TI Header';
    exit;
    end;
  //
  if NumRead <> 128 then begin
    statMainInfoMessage.Caption := 'Error Reading TI Header';
    CloseFile(FromF);
    exit;
    end;
  //
  try
    //
    for i := 1 to CurrentFile.Header.Sectors do begin
      statMainInfoMessage.Caption := 'Reading TI Sector #...' + inttostr(i);
      try
        BlockRead(FromF, CurrentFile.SectorList[i - 1].Contents,
          SizeOf(CurrentFile.SectorList[i - 1].Contents), NumRead);
        CurrentFile.SectorList[i - 1].EndOffset := NumRead;
      except
        //
      end; //try
      end; //for
  //
  finally
    CloseFile(FromF);
  end;     //try
  //
  // these checks are kind of kludgy, so I improve the odds of being
  //    accurate by submitting only likely file types
  //
  if CurrentFile.Header.ProgType then
    CurrentFile.BasicFlag := CheckforBasic;
  if not CurrentFile.Header.ProgType then begin
    CurrentFile.CompanionFlag := CheckforComp;
    CurrentFile.TIWriterFlag := CheckforWriter;
    end;
  //
  ckBASIC.Checked := CurrentFile.BASICFlag;
  ckTIWriter.Checked := CurrentFile.TIWriterFlag;
  ckCompanion.Checked := CurrentFile.CompanionFlag;
  //
  CurrentFile.SectorsRead := true;
  statMainInfoMessage.Caption := 'Read ' + inttostr(i - 1) + ' Sectors';
  //
  if CurrentFile.SectorsRead then begin
    //
    ckBASIC.Enabled := CurrentFile.SectorsRead and
        (Copy(CurrentFile.Header.FileType, 1, 5) = 'IMAGE');
    ckCompanion.Enabled := CurrentFile.SectorsRead and
        (Copy(CurrentFile.Header.FileType, 1, 5) = 'IV254');
    ckTIWriter.Enabled := CurrentFile.SectorsRead and
        (Copy(CurrentFile.Header.FileType, 1, 5) = 'DV 80');
    //
    DisplayingDiskImage := False;
    DisplayingDiskFile := False;
    DisplayingFile := True;
    DisplayTISector(0, CurrentFile.SectorList[0].Contents);
    PagesTI.ActivePage := tabFileInfo;
    end;
end;

//
//  Other Subroutines
//

procedure TfrmMain.ClearCurrentDiskette;
begin
  CurrentDisk.PCName := '';
    CurrentDisk.Info.Diskname := 'NO TI DISK';
    CurrentDisk.Info.SectorCount := 0;
    CurrentDisk.Info.SectorsPerTrack := 0;
    CurrentDisk.Info.DSRMark := '';
    CurrentDisk.Info.DiskProtected := false;
    CurrentDisk.Info.TracksPerSide := 0;
    CurrentDisk.Info.Sides := 0;
    CurrentDisk.Info.Density := 0;
  //CurrentDisk.SectorList := nil;
  //CurrentDisk.Directory := nil;
  CurrentDisk.SectorsRead := False;
    //CurrentDisk.Info.Bitmap := nil; //<-- Uncomment to get errors!
end;

procedure TfrmMain.ClearCurrentFile;
begin
  //CurrentFile.SectorList := nil;
    CurrentFile.Header.HeaderString := 'NOT  TI';
    CurrentFile.Header.Sectors := 0;
    CurrentFile.Header.RecCount := 0;
    CurrentFile.Header.StatusFlag := 0;
    CurrentFile.Header.FileType := '';
    CurrentFile.Header.DataSize := '';
    CurrentFile.Header.DataType := '';
    CurrentFile.Header.ProtectBit := false;
    CurrentFile.Header.RecSize := 0;
    CurrentFile.Header.ProgType := False;
    CurrentFile.Header.RecordsPerSec := 0;
    CurrentFile.Header.EOFOffset := 0;
  CurrentFile.Name := '';
  CurrentFile.PCName := '';
  CurrentFile.StandAlone := False;
  CurrentFile.SectorsRead := False;
  CurrentFile.BasicFlag := False;
  CurrentFile.CompanionFlag := False;
  CurrentFile.TIWriterFlag := False;
  CurrentFile.FormattedFlag := False;
end;

procedure TfrmMain.ClearDisketteInfo;
begin
  edDiskName.text := 'NO TI DISK';
  ckProtected.Checked := False;
  edSectorCount.Value := 0;
  edSectorsPerTrack.Value := 0;
  edTracksPerSide.Value := 0;
  edSides.Value := 0;
  edDensity.Value := 0;
  ckProtected.checked := False;
  //
  lbCatalog.Clear;
end;

procedure TfrmMain.ClearTIFILESFileInfo;
var i, j: integer;
begin
  //
  //  Reset all displays
  //
  ckBASIC.Checked := false;
  ckCompanion.Checked := false;
  ckTIWriter.Checked := false;
  //
  for i := 1 to GridSectors.RowCount - 1 do
    for j := 1 to gridSectors.ColCount - 1 do
      GridSectors.cells[j, i] := '';
  //
  for i := 0 to GridHeader.RowCount - 1 do
    for j := 0 to gridHeader.ColCount - 1 do
      GridHeader.cells[j, i] := '';
  //
  //  Reset TIHEADER display
  //
  ckTIFilesHeader.Checked := false;
  ckBASIC.Enabled := False;
  ckBASIC.Checked := False;
  ckCompanion.Enabled := False;
  ckCompanion.Checked := False;
  ckTIWriter.Enabled := False;
  ckTIWriter.Checked := False;
  //
  edEOFOffset.Value := 0;
  edEOFOffsetHex.Text := '0';
  edRecCount.Visible := False;
  edRecCountHex.Visible := False;
  edRecPerSector.Visible := False;
  edRecPerSectorHex.Visible := False;
  edRecSize.Visible := False;
  edRecSizeHex.Visible := False;
  edTISectorCount.Value := 0;
  edTISectorCountHex.Text := '0';
  lblTIFileName.Caption := 'NO TI FILE';
  lblTIFileType.caption := '';
  rgDataType.Visible := False;
  rgDataSize.Visible := False;
  label1.Visible := False;
  label2.Visible := False;
  label3.Visible := False;
end;

procedure TfrmMain.ClearMemos;
begin
  memFormatted.Clear;
  memRawText.Clear;
end;

//
//  Action Management
//

procedure TfrmMain.actionFormattedUpdate(Action: TBasicAction;
  var Handled: Boolean);
begin
  actSaveAsText.Enabled := CurrentFile.FormattedFlag;
  actPrintText.Enabled := CurrentFile.FormattedFlag;
end;

procedure TfrmMain.actionMainUpdate(Action: TBasicAction;
  var Handled: Boolean);
begin
   actFormatCompanion.Enabled :=
     (not CurrentFile.Header.ProgType)
     and CurrentFile.CompanionFlag;
   //
   actDisplayText.Enabled :=
     (((not CurrentFile.Header.ProgType)
     and (CurrentFile.Header.DataType = 'DIS')
     and (CurrentFile.SectorsRead)
     and (not CurrentFile.CompanionFlag)));
   //
   actParseBasic.Enabled :=
     ((CurrentFile.Header.ProgType)
     and (CurrentFile.SectorsRead)
     and (CurrentFile.BasicFlag));
   //
   actPrintText.Enabled := CurrentFile.FormattedFlag;
   actSaveAsText.Enabled := CurrentFile.FormattedFlag;
   //
   actCatalog.Enabled := CurrentDisk.SectorsRead;
   actDisplayFile.Enabled := CurrentDisk.SectorsRead;
   actExtractFile.Enabled :=
     (not CurrentFile.StandAlone) and CurrentFile.SectorsRead;
   //
   actListNarrow.Enabled :=
     ((CurrentFile.Header.ProgType)
     and (CurrentFile.FormattedFlag)
     and (CurrentFile.BasicFlag));
   //
   FontEdit.Enabled := memFormatted.SelLength > 0;
   //
   //  some interface stuff
   //
   spinFontSize.Enabled := cbFonts.Enabled;
   cbFonts.Enabled := CurrentFile.SectorsRead;
end;

//
//  various GUI interface handlers
//

//
//  Change Sector Number response
//

procedure TfrmMain.spinSectorUpClick(Sender: TObject);
begin
  if strtoint('$' + edSectorNumHex.Text) < MaxSectors
    then begin
    inc(CurrentSector);
    //
    if DisplayingFile then
     DisplayTISector(CurrentSector,
       CurrentFile.SectorList[CurrentSector].Contents)
    else if DisplayingDiskImage then
     DisplayTISector(CurrentSector,
       CurrentDisk.SectorList[CurrentSector].Contents);
    //
    edSectorNumHex.Text := inttohex(CurrentSector, 2);
    edSectorNum.Value := strtoint('$' + edSectorNumHex.Text);
    end;
end;

procedure TfrmMain.spinSectorDownClick(Sender: TObject);
begin
  if strtoint('$' + edSectorNumHex.Text) > 0
    then begin
    dec(CurrentSector);
    //
    if DisplayingFile then
     DisplayTISector(CurrentSector,
       CurrentFile.SectorList[CurrentSector].Contents)
    else if DisplayingDiskImage then
     DisplayTISector(CurrentSector,
       CurrentDisk.SectorList[CurrentSector].Contents);
    //
    edSectorNumHex.Text := inttohex(CurrentSector, 2);
    edSectorNum.Value := strtoint('$' + edSectorNumHex.Text);
    end;
end;

procedure TfrmMain.edSectorNumExit(Sender: TObject);
begin
  if strtoint(edSectorNum.Text) < MaxSectors
    then begin
    CurrentSector := strtoint(edSectorNum.Text);
    edSectorNumHex.Text := inttohex(CurrentSector, 2);
    //
    if DisplayingFile then
     DisplayTISector(CurrentSector,
       CurrentFile.SectorList[CurrentSector].Contents)
    else if DisplayingDiskImage then
     DisplayTISector(CurrentSector,
       CurrentDisk.SectorList[CurrentSector].Contents);
    end;
end;

procedure TfrmMain.gridSectorsOriginalClick(Sender: TObject);
begin
  if gridSectors.Col < 9 then begin
    edByteNumHex.Text :=
      inttohex((gridSectors.Col - 1) + (8 * (gridSectors.Row - 1)), 2);
    //
    edContent.text := inttostr(strtoint('$'
     + gridSectors.Cells[GridSectors.Col, GridSectors.Row]));
    edContentChar.text := CHR(strtoint('$'
     + gridSectors.Cells[GridSectors.Col, GridSectors.Row]));
    end
  else begin
    edByteNumHex.Text :=
      inttohex((gridSectors.Col - 9) + (8 * (gridSectors.Row - 1)), 2);
    //
    edContentChar.text := gridSectors.Cells[GridSectors.Col,
       GridSectors.Row];
    try
     edContent.text := inttostr(ord(edContentChar.text[1])); except end;
    end;
  //
  edByteNum.Value := strtoint('$' + edByteNumHex.Text)
end;

procedure TfrmMain.NewFonts(Sender: TObject);
var i: integer;
begin
   memRawText.Font.Name := cbFonts.FontName;
   memRawText.Font.Size := spinFontSize.AsInteger;
   //
   memFormatted.Font.Name := cbFonts.FontName;
   memFormatted.font.Size := spinFontSize.AsInteger;
   gridHeader.Font.Name :=  cbFonts.FontName;
   gridHeader.Font.Size := spinFontSize.AsInteger;
   gridSectors.Font.Name :=  cbFonts.FontName;
   gridSectors.Font.Size := spinFontSize.AsInteger;
   for i := 0 to gridSectors.ColCount - 1 do begin
     gridSectors.Columns.Items[i].Font.Name := cbFonts.FontName;
     gridSectors.Columns.items[i].FOnt.Size := spinFontSize.AsInteger;
     end;
   lbCatalog.Font.Name := cbFonts.FontName;
   lbCatalog.Font.Size := spinFontSIze.AsInteger;
end;

//
//  RichEdit Routines
//

procedure TfrmMain.FontEditAccept(Sender: TObject);
begin
   memFormatted.SelAttributes.Name := FontEdit.Dialog.Font.Name;
   memFormatted.SelAttributes.Size := FontEdit.Dialog.Font.Size;
   memFormatted.SelAttributes.Style := FontEdit.Dialog.Font.Style;
end;

procedure TfrmMain.actPrintTextAccept(Sender: TObject);
var  LM, RM, TM, BM: integer;
begin
  //
  //  we may need these as program variables someday
  //
  LM := 1000;
  TM := 1000;
  RM := 1000;
  BM := 1000;
  //
  //  do this here every time because I'm paranoid
  //
  PrintMargins.TopLeft := Point(LM, TM);
  PrintMargins.BottomRight := Point(RM, BM);     // These are treated as offsets
                                                 // from the right and bottom
                                                 // by the internal margin
                                                 // routine
  SetRichEdMargins(memFormatted, PrintMargins);
  //
  memFormatted.Print(CurrentFile.PCName);
end;

procedure TfrmMain.actCatalogExecute(Sender: TObject);
var i, prow, printx, printy, colwid, HalfFactor: integer;
    msg: string;
begin
with Printer do begin
  //    X = Horizontal position
  //    Y = Vertical Position
  BeginDoc;
  //
  //  DiskName and Info
  //
  canvas.font.name := lbCatalog.Font.Name;
  canvas.font.size := 14;
  canvas.Font.Style := canvas.Font.Style + [fsBold];
  canvas.font.color := clNavy;
  prow := pageheight div 30;
  printy := prow;
  colwid := pagewidth div 5;
  msg := 'Disk Image Name: ' + CurrentDisk.Info.Diskname;
  canvas.textout(colwid, printy, msg);
  //
  printy := printy + canvas.textheight(msg);
  canvas.font.color := clBlack;
  canvas.font.size := 10;
  canvas.Font.Style := canvas.Font.Style - [fsBold];
  if CurrentDisk.Info.Sides = 2 then msg := 'DS' else msg := 'SS';
  if CurrentDisk.Info.Density = 2 then
    msg := msg + 'DD' else msg := msg + 'SD';
  msg := msg +
   ' -- Total Sectors: ' + Inttostr(CurrentDisk.Info.SectorCount)
    + '  --  Total Files: ' + Inttostr(CurrentDisk.Info.FileCount);
  canvas.textout(colwid, printy, msg);
  printy := printy + canvas.textheight(msg);
  msg := '================================================';
  canvas.textout(colwid, printy, msg);
  printy := printy + canvas.textheight(msg);
  msg := 'name      size  type p    name      size  type p';
  canvas.textout(colwid, printy, msg);
  printy := printy + canvas.textheight(msg);
  msg := '------------------------------------------------';
  canvas.textout(colwid, printy, msg);
  //
  //  Catalog TODO  odd numbers fail?
  //
  i := 0;
  HalfFactor := (lbCatalog.Items.Count - 1) div 2;
  while i <= (HalfFactor - (lbCatalog.Items.Count mod 2)) do begin
    printy := printy + canvas.textheight(msg);
    msg := lbCatalog.ItemPart(i, 0) +
           #32 + format('%3d', [strtoint(lbCatalog.ItemPart(i, 1))]) +
           #32 + lbCatalog.ItemPart(i, 2) +
           #32 + #32 + #32 + #32 + lbCatalog.ItemPart(i + HalfFactor +
                    1 - (lbCatalog.Items.Count mod 2), 0) +
           #32 + format('%3d', [strtoint(lbCatalog.ItemPart(i + HalfFactor +
                    1 - (lbCatalog.Items.Count mod 2), 1))]) +
           #32 + lbCatalog.ItemPart(i + HalfFactor +
                    1 - (lbCatalog.Items.Count mod 2), 2);
    canvas.textout(colwid, printy, msg);
    inc(i);
    end;
  //
  if (lbCatalog.Items.Count mod 2) = 1 then begin
    printy := printy + canvas.textheight(msg);
    msg := stringofchar(#32, 26)
           + lbCatalog.ItemPart(lbCatalog.Items.Count - 1, 0) +
           #32 + format('%3d',
             [strtoint(lbCatalog.ItemPart(lbCatalog.Items.Count - 1, 1))]) +
           #32 + lbCatalog.ItemPart(lbCatalog.Items.Count - 1, 2);
    canvas.textout(colwid, printy, msg);
    end;
  //
  EndDoc;
  End;
end;

procedure TfrmMain.actSaveAsTextAccept(Sender: TObject);
begin
   memFormatted.Lines.SaveToFile(actSaveAsText.Dialog.FileName);
   //memFormatted.OutputFormat := ofText;    //FIXME
end;

procedure TfrmMain.actSaveasRTFExecute(Sender: TObject);
begin
   //memFormatted.OutputFormat := ofRTFNoObjs;
   SaveDialog.Filter := 'RTF Files|*.rtf|All Files|*.*';
   SaveDialog.DefaultExt := '*.rtf';
   SaveDialog.Title := 'Save formatted text as RTF';
   if SaveDialog.Execute
     then memFormatted.Lines.SaveToFile(SaveDialog.FileName);
end;

procedure TfrmMain.actListNarrowExecute(Sender: TObject);
//********************************************
//*                                          *
//*  Why anyone would want to do this        *
//*    [and why I wrote it]                  *
//*    I'll never understand.......          *
//*                                          *
//********************************************
var NewCols: TStrings;
    i, j: integer;
begin
   NewCols := TStringList.Create;
   try
   for i := 0 to memFormatted.Lines.count do begin
     for j := 0 to (length(memFormatted.Lines[i]) div 28) do
        NewCols.Add(Copy(memFormatted.Lines[i], j * 28, 28));
     end;
   memFormatted.Lines.Assign(NewCols);
   finally
   NewCols.Free;
   end;
end;

procedure TfrmMain.ckBasicClick(Sender: TObject);
begin
   If CurrentFile.SectorsRead then
     CurrentFile.BASICFlag := not CurrentFile.BasicFlag;
end;

procedure TfrmMain.ckCompanionClick(Sender: TObject);
begin
   If CurrentFile.SectorsRead then
     CurrentFile.CompanionFlag := not CurrentFile.CompanionFlag;
end;

procedure TfrmMain.ckTIWriterClick(Sender: TObject);
begin
   If CurrentFile.SectorsRead then
     CurrentFile.TIWriterFlag :=
       not CurrentFile.TIWriterFlag;
end;

procedure TfrmMain.dlgAboutCustomClick(Sender: TObject);
begin
  Application.HelpCommand(HELP_FINDER, 0);
end;

procedure TfrmMain.gridDiskBitMapDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
//*************************************************
//* main work is done by UsedSegment routine      *
//*************************************************
var Segnumber: integer;
begin
  if not ResizingGrid then begin
    SegNumber := ARow * gridDiskBitMap.ColCount + ACol;
    with (Sender as TDrawGrid) do begin
      with Canvas do begin
        //
        //  red for used, aqua for available...
        //
        if UsedSegment(Segnumber) then Brush.Color := clRed
            else Brush.Color := clAqua;
        FillRect(Rect);
        end;  // with Canvas
      end;  //eith Sender
    end; // if not Resizing
end;

procedure TfrmMain.lbCatalogDblClick(Sender: TObject);
begin
  actDisplayFileExecute(Self);
end;

procedure TfrmMain.actOpenBeforeExecute(Sender: TObject);
begin
  ActOpen.Dialog.Filter := 'All Files|*.*';
end;

procedure TfrmMain.FileListBox1DblClick(Sender: TObject);
var i,j: integer;
    ss: string;
begin
    screen.Cursor := crHourGlass;
    memFormatted.Clear;
    statMainInfoMessage.Caption := 'Stage One';
    //
    //  Reset all displays, records and arrays
    //
    statMainInfoMessage.Caption := 'Clearing File Info';
    ClearTIFilesFileInfo;
    statMainInfoMessage.Caption := 'Clearing Disk Display';
    ClearDisketteInfo;
    statMainInfoMessage.Caption := 'Clearing Disk Array';
    ClearCurrentDiskette;
    //
    //  Get valid disk info
    //
    statMainInfoMessage.Caption := 'Opening Image.......';
    if ReadTIImage(FileListBox1.FileName) then begin
      edDiskName.text := CurrentDisk.Info.DiskName;
      ckProtected.Checked := CurrentDisk.Info.DiskProtected;
      edSectorCount.Value := CurrentDisk.Info.SectorCount;
      edSectorsPerTrack.Value := CurrentDisk.Info.SectorsPerTrack;
      edTracksPerSide.Value := CurrentDisk.Info.TracksPerSide;
      edSides.Value := CurrentDisk.Info.Sides;
      edDensity.Value := CurrentDisk.Info.Density;
      //
      // Display Directory
      //
      lbCatalog.Clear;
      statMainInfoMessage.Caption := '   Building catalog...';
      for i := 0 to CurrentDisk.Info.FileCount - 1 do begin
         // Construct Starting Segment List
         j := 0;  SS := '';
         repeat
            SS := SS
             + inttostr(CurrentDisk.Directory[i].SegmentList[j].StartSeg)
             + ','
             + inttostr(CurrentDisk.Directory[i].SegmentList[j].EndOffset);
            inc(j);
            SS := SS + #32;
         until ord(CurrentDisk.Directory[i].SegmentList[j].EndOffset) = 0;
         //
         lbCatalog.AddLine
           ([CurrentDisk.Directory[i].FileName,
             inttostr(CurrentDisk.Directory[i].SectorCount),
             DecodeType(CurrentDisk.Directory[i].StatusFlag,
                        CurrentDisk.Directory[i].Recordlength),
             SS,
             inttostr(CurrentDisk.Directory[i].Address)]);
         end;
      //
      statMainInfoMessage.caption :=
         inttostr(CurrentDisk.Info.SectorCount) + ' sectors - '
         + inttostr(CurrentDisk.Info.FileCount) + ' files';
      statMainInfoMessage2.Caption := OpenImage.FileName;
      statSecondaryInfoMessage1.Caption := 'TI Diskette Image in Memory';
      //
      PagesTI.ActivePage := tabImageInfo;
      //
      DisplayTISector(0, CurrentDisk.SectorList[0].Contents);
      end;
    screen.Cursor := crDefault;
end;

end.     //The End
